import os

from brian2 import *

HEADER = '''<?xml version="1.0" encoding="iso-8859-1"?>
<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 20001102//EN"
 "http://www.w3.org/TR/2000/CR-SVG-20001102/DTD/svg-20001102.dtd">

<svg height="3em" viewBox="{minx} {miny} {width} {height}"
    xmlns="http://www.w3.org/2000/svg"
    xmlns:xlink="http://www.w3.org/1999/xlink/">
'''

# For now, we do not bother showing the actual size of the Soma
SPHERE = '''
    <defs>
      <radialGradient id="blueSphere"
                      cx="0" cy="0" r="100%" fx="-50%" fy="-50%">
        <stop offset="0%" stop-color="white" />
        <stop offset="75%" stop-color="blue" />
        <stop offset="100%" stop-color="darkblue" />
      </radialGradient>
    </defs>
    <circle fill="url(#blueSphere)" cx="0" cy="0" r="5" />
'''

CYLINDER = '''
<g transform="translate({startx},0)">
   <ellipse cx="0"  cy="{center}"
            rx="0.5" ry="{radius}"
            stroke="blue" stroke-width="1"
            fill="blue"/>
   <rect x="0" y="{starty}" width="{length}" height="{diameter}" fill="blue"
    stroke="blue" stroke-width="1"/>
   <ellipse cx="{length}" cy="{center}"
            rx="0.5"  ry="{radius}"
            stroke="darkblue" fill="darkblue" stroke-width="1"/>
</g>
'''

FRUSTRUM = '''
<g transform="translate({startx},0)">
   <ellipse cx="0"  cy="{center}"
            rx="0.5" ry="{radius1}"
            stroke="blue" stroke-width="1"
            fill="blue"/>
   <polygon points="0 {starty1}, 0 {starty2}, {length} {endy2}, {length} {endy1}"
    fill="blue" stroke="blue" stroke-width="1"/>
   <ellipse cx="{length}" cy="{center}"
            rx="0.5"  ry="{radius2}"
            stroke="darkblue" fill="darkblue" stroke-width="1"/>
</g>
'''

PARENT_CYLINDER = '''
<defs>
<linearGradient id="gradient_to_transparent_grey" x1="0" y1="0" x2="100%" y2="0">
    <stop offset="0" stop-color="white" stop-opacity="0"/>
    <stop offset="0.5" stop-color="#222222"  stop-opacity="1"/>
    <stop offset="1" stop-color="#222222"  stop-opacity="1"/>
</linearGradient>
</defs>
<g transform="translate(-{length},0)">
   <rect x="-3" y="{starty}" width="{length}" height="{diameter}" fill="url(#gradient_to_transparent_grey)"
    stroke="url(#gradient_to_transparent_grey)" stroke-width="1"/>
   <g transform="translate(-3,0)">
   <ellipse cx="{length}" cy="{center}"
            rx="0.5"  ry="{radius}"
            stroke="black" fill="black" stroke-width="1"/>
   </g>
</g>
'''

CONNECTION = '''
<line x1="{start}" x2="{end}" y1="{center}" y2="{center}" stroke="black" stroke-width="1"/>
'''
FOOTER = '''
</svg>
'''

def to_svg(morphology):
    if isinstance(morphology, Soma):
        return HEADER.format(minx='-5', miny='-5', width='10', height='10') + SPHERE.format(radius=morphology.diameter[0]/2/um) + FOOTER
    if isinstance(morphology, Cylinder):
        summed_length = 0*um
        elements = []
        center = max([7.5*um, max(morphology.end_diameter)/2])
        minx = -2.5
        if morphology.parent is not None:
            elements.append(PARENT_CYLINDER.format(starty=(center-morphology.parent.end_diameter[-1]/2)/um,
                                                   length=morphology.parent.length[-1]/um,
                                                   center=center/um,
                                                   diameter=morphology.parent.end_diameter[-1]/um,
                                                   radius=morphology.parent.end_diameter[-1]/2/um))
            minx -= morphology.parent.length[-1]/um
        for idx, (diameter, length) in enumerate(zip(morphology.diameter, morphology.length)):
            elements.append(CONNECTION.format(start=(summed_length-3*um)/um,
                                              end=summed_length/um,
                                              center=center/um))
            elements.append(CYLINDER.format(startx=summed_length/um,
                                            starty=(center-diameter/2)/um,
                                            radius=diameter/2/um,
                                            diameter=diameter/um,
                                            length=length/um,
                                            center=center/um))
            summed_length += length + 3*um
        return HEADER.format(minx=minx, miny=-1, width=summed_length/um+0.5-minx, height=center*2/um+2) + ('\n'.join(elements)) + FOOTER
    elif isinstance(morphology, Section):
        summed_length = 0*um
        elements = []
        center = max([7.5*um, max(morphology.end_diameter)/2])
        minx = -5
        if morphology.parent is not None:
                elements.append(PARENT_CYLINDER.format(starty=(center-morphology.parent.end_diameter[-1]/2)/um,
                                                       length=morphology.parent.length[-1]/um,
                                                       center=center/um,
                                                       diameter=morphology.parent.end_diameter[-1]/um,
                                                       radius=morphology.parent.end_diameter[-1]/2/um))
                minx -= morphology.parent.length[-1]/um
        for idx, (start_diameter, end_diameter, length) in enumerate(zip(morphology.start_diameter, morphology.end_diameter, morphology.length)):
            elements.append(CONNECTION.format(start=(summed_length-3*um)/um,
                                              end=summed_length/um,
                                              center=center/um))
            elements.append(FRUSTRUM.format(startx=summed_length/um,
                                            starty1=(center-start_diameter/2)/um,
                                            starty2=(center+start_diameter/2)/um,
                                            endy1=(center-end_diameter/2)/um,
                                            endy2=(center+end_diameter/2)/um,
                                            radius1=start_diameter/2/um,
                                            radius2=end_diameter/2/um,
                                            length=length/um,
                                            center=center/um))
            summed_length += length + 3*um
        return HEADER.format(minx=minx, miny=-1, width=summed_length/um+1.5-minx,
                             height=center*2/um+2) + ('\n'.join(elements)) + FOOTER
    else:
        raise NotImplementedError()
if __name__ == '__main__':
    dirname = os.path.dirname(__file__)
    PATH = os.path.join(dirname, '..', '..', '..', 'docs_sphinx', 'user', 'images')
    root = Cylinder(n=1, diameter=15*um, length=10*um)
    root.cyl = Cylinder(n=5, diameter=10*um, length=50*um)
    root.sec = Section(n=5, diameter=[15, 5, 10, 5, 10, 5]*um,
                       length=[10, 20, 5, 5, 10]*um)
    for filename, morpho in [('soma.svg', Soma(diameter=30*um)),
                             ('cylinder.svg', root.cyl),
                             ('section.svg', root.sec)]:
        with open(os.path.join(PATH, filename), 'w') as f:
            print(filename)
            f.write(to_svg(morpho))



"""
Cython automatic extension builder/manager

Inspired by IPython's Cython cell magics, see:
https://github.com/ipython/ipython/blob/master/IPython/extensions/cythonmagic.py
"""

import glob
import hashlib
import importlib.util
import os
import shutil
import sys
import time
from distutils.command.build_ext import build_ext
from distutils.core import Distribution, Extension

import numpy

try:
    import Cython
    import Cython.Build as Cython_Build
    import Cython.Compiler as Cython_Compiler

    try:
        from Cython.Utils import get_cython_cache_dir as base_cython_cache_dir
    except ImportError:
        from Cython.Build.Cache import (
            get_cython_cache_dir as base_cython_cache_dir,  # Cython 3.1, see cython/cython#6090
        )
except ImportError:
    Cython = None

from brian2.core.preferences import prefs
from brian2.utils.filelock import FileLock
from brian2.utils.logger import get_logger, std_silent
from brian2.utils.stringtools import deindent

__all__ = ["cython_extension_manager"]

logger = get_logger(__name__)


def get_cython_cache_dir():
    cache_dir = prefs.codegen.runtime.cython.cache_dir
    if cache_dir is None and Cython is not None:
        cache_dir = os.path.join(base_cython_cache_dir(), "brian_extensions")
    return cache_dir


def get_cython_extensions():
    return {
        ".pyx",
        ".pxd",
        ".pyd",
        ".cpp",
        ".c",
        ".so",
        ".o",
        ".o.d",
        ".lock",
        ".dll",
        ".obj",
        ".exp",
        ".lib",
    }


class CythonExtensionManager:
    def __init__(self):
        self._code_cache = {}

    def create_extension(
        self,
        code,
        force=False,
        name=None,
        define_macros=None,
        include_dirs=None,
        library_dirs=None,
        runtime_library_dirs=None,
        extra_compile_args=None,
        extra_link_args=None,
        libraries=None,
        compiler=None,
        sources=None,
        owner_name="",
    ):
        if sources is None:
            sources = []
        self._simplify_paths()

        if Cython is None:
            raise ImportError("Cython is not available")

        code = deindent(code)

        lib_dir = get_cython_cache_dir()
        if "~" in lib_dir:
            lib_dir = os.path.expanduser(lib_dir)
        try:
            os.makedirs(lib_dir)
        except OSError:
            if not os.path.exists(lib_dir):
                raise OSError(
                    f"Couldn't create Cython cache directory '{lib_dir}', try setting"
                    " the cache directly with prefs.codegen.runtime.cython.cache_dir."
                )

        numpy_version = ".".join(
            numpy.__version__.split(".")[:2]
        )  # Only use major.minor version
        # avoid some issues when manually switching compilers
        CC = os.environ.get("CC", None)
        CXX = os.environ.get("CXX", None)
        key = (
            code,
            sys.version_info,
            sys.executable,
            Cython.__version__,
            numpy_version,
            CC,
            CXX,
        )

        if force:
            # Force a new module name by adding the current time to the
            # key which is hashed to determine the module name.
            key += (time.time(),)  # Note the trailing comma (this is a tuple)

        if key in self._code_cache:
            return self._code_cache[key]

        if name is not None:
            module_name = name  # py3compat.unicode_to_str(args.name)
        else:
            module_name = (
                f"_cython_magic_{hashlib.md5(str(key).encode('utf-8')).hexdigest()}"
            )
        if owner_name:
            logger.diagnostic(f'"{owner_name}" using Cython module "{module_name}"')

        module_path = os.path.join(lib_dir, module_name + self.so_ext)

        if prefs["codegen.runtime.cython.multiprocess_safe"]:
            lock = FileLock(os.path.join(lib_dir, f"{module_name}.lock"))
            with lock:
                module = self._load_module(
                    module_path,
                    define_macros=define_macros,
                    include_dirs=include_dirs,
                    library_dirs=library_dirs,
                    extra_compile_args=extra_compile_args,
                    extra_link_args=extra_link_args,
                    libraries=libraries,
                    code=code,
                    lib_dir=lib_dir,
                    module_name=module_name,
                    runtime_library_dirs=runtime_library_dirs,
                    compiler=compiler,
                    key=key,
                    sources=sources,
                )
            return module
        else:
            return self._load_module(
                module_path,
                define_macros=define_macros,
                include_dirs=include_dirs,
                library_dirs=library_dirs,
                extra_compile_args=extra_compile_args,
                extra_link_args=extra_link_args,
                libraries=libraries,
                code=code,
                lib_dir=lib_dir,
                module_name=module_name,
                runtime_library_dirs=runtime_library_dirs,
                compiler=compiler,
                key=key,
                sources=sources,
            )

    @property
    def so_ext(self):
        """The extension suffix for compiled modules."""
        try:
            return self._so_ext
        except AttributeError:
            self._so_ext = self._get_build_extension().get_ext_filename("")
            return self._so_ext

    def _clear_distutils_mkpath_cache(self):
        """clear distutils mkpath cache

        prevents distutils from skipping re-creation of dirs that have been removed
        """
        try:
            from distutils.dir_util import _path_created
        except ImportError:
            pass
        else:
            _path_created.clear()

    def _get_build_extension(self, compiler=None):
        self._clear_distutils_mkpath_cache()
        dist = Distribution()
        config_files = dist.find_config_files()
        try:
            config_files.remove("setup.cfg")
        except ValueError:
            pass
        dist.parse_config_files(config_files)
        build_extension = build_ext(dist)
        if compiler is not None:
            build_extension.compiler = compiler
        build_extension.finalize_options()
        return build_extension

    def _load_module(
        self,
        module_path,
        define_macros,
        include_dirs,
        library_dirs,
        extra_compile_args,
        extra_link_args,
        libraries,
        code,
        lib_dir,
        module_name,
        runtime_library_dirs,
        compiler,
        key,
        sources,
    ):
        have_module = os.path.isfile(module_path)

        if not have_module:
            if define_macros is None:
                define_macros = []
            if include_dirs is None:
                include_dirs = []
            if library_dirs is None:
                library_dirs = []
            if runtime_library_dirs is None:
                runtime_library_dirs = []
            if extra_compile_args is None:
                extra_compile_args = []
            if extra_link_args is None:
                extra_link_args = []
            if libraries is None:
                libraries = []

            c_include_dirs = include_dirs
            if "numpy" in code:
                import numpy

                c_include_dirs.append(numpy.get_include())

            # TODO: We should probably have a special folder just for header
            # files that are shared between different codegen targets
            import brian2.synapses as synapses

            synapses_dir = os.path.dirname(synapses.__file__)
            c_include_dirs.append(synapses_dir)

            pyx_file = os.path.join(lib_dir, f"{module_name}.pyx")
            # ignore Python 3 unicode stuff for the moment
            # pyx_file = py3compat.cast_bytes_py2(pyx_file, encoding=sys.getfilesystemencoding())
            # with io.open(pyx_file, 'w', encoding='utf-8') as f:
            #    f.write(code)
            with open(pyx_file, "w") as f:
                f.write(code)

            for source in sources:
                if not source.lower().endswith(".pyx"):
                    raise ValueError(
                        "Additional Cython source files need to have an .pyx ending"
                    )
                # Copy source and header file (if present) to library directory
                shutil.copyfile(source, os.path.join(lib_dir, os.path.basename(source)))
                name_without_ext = os.path.splitext(os.path.basename(source))[0]
                header_name = f"{name_without_ext}.pxd"
                if os.path.exists(os.path.join(os.path.dirname(source), header_name)):
                    shutil.copyfile(
                        os.path.join(os.path.dirname(source), header_name),
                        os.path.join(lib_dir, header_name),
                    )
            final_sources = [
                os.path.join(lib_dir, os.path.basename(source)) for source in sources
            ]
            extension = Extension(
                name=module_name,
                sources=[pyx_file],
                define_macros=define_macros,
                include_dirs=c_include_dirs,
                library_dirs=library_dirs,
                runtime_library_dirs=runtime_library_dirs,
                extra_compile_args=extra_compile_args,
                extra_link_args=extra_link_args,
                libraries=libraries,
                language="c++",
            )
            build_extension = self._get_build_extension(compiler=compiler)
            try:
                opts = dict(
                    quiet=True,
                    annotate=False,
                    force=True,
                )
                # suppresses the output on stdout
                with std_silent():
                    build_extension.extensions = Cython_Build.cythonize(
                        [extension] + final_sources, **opts
                    )

                    build_extension.build_temp = os.path.dirname(pyx_file)
                    build_extension.build_lib = lib_dir
                    build_extension.run()
                    if prefs["codegen.runtime.cython.delete_source_files"]:
                        # we can delete the source files to save disk space
                        cpp_file = os.path.join(lib_dir, f"{module_name}.cpp")
                        try:
                            os.remove(pyx_file)
                            os.remove(cpp_file)
                            temp_dir = os.path.join(
                                lib_dir,
                                os.path.dirname(pyx_file)[1:],
                                f"{module_name}.*",
                            )
                            for fname in glob.glob(temp_dir):
                                os.remove(fname)
                        except OSError as ex:
                            logger.debug(
                                "Deleting Cython source files failed with error:"
                                f" {str(ex)}"
                            )

            except Cython_Compiler.Errors.CompileError:
                return
        # Temporarily insert the Cython directory to the Python path so that
        # code importing from an external module that was declared via
        # sources works
        sys.path.insert(0, lib_dir)
        spec = importlib.util.spec_from_file_location(module_name, module_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        sys.path.pop(0)
        self._code_cache[key] = module
        return module

    def _simplify_paths(self):
        if "lib" in os.environ:
            os.environ["lib"] = simplify_path_env_var(os.environ["lib"])
        if "include" in os.environ:
            os.environ["include"] = simplify_path_env_var(os.environ["include"])


cython_extension_manager = CythonExtensionManager()


def simplify_path_env_var(path):
    allpaths = path.split(os.pathsep)
    knownpaths = set()
    uniquepaths = []
    for p in allpaths:
        if p not in knownpaths:
            knownpaths.add(p)
            uniquepaths.append(p)
    return os.pathsep.join(uniquepaths)



"""
Simplify and optimise sequences of statements by rewriting and pulling out loop invariants.
"""

import ast
import copy
import itertools
from collections import OrderedDict
from functools import reduce

from brian2.core.functions import DEFAULT_CONSTANTS, DEFAULT_FUNCTIONS
from brian2.core.preferences import prefs
from brian2.core.variables import AuxiliaryVariable
from brian2.parsing.bast import (
    BrianASTRenderer,
    brian_ast,
    brian_dtype_from_dtype,
    dtype_hierarchy,
)
from brian2.parsing.rendering import NodeRenderer
from brian2.utils.stringtools import get_identifiers, word_substitute

from .statements import Statement

# Default namespace has all the standard functions and constants in it
defaults_ns = {k: v.pyfunc for k, v in DEFAULT_FUNCTIONS.items()}
defaults_ns.update({k: v.value for k, v in DEFAULT_CONSTANTS.items()})


__all__ = ["optimise_statements", "ArithmeticSimplifier", "Simplifier"]


def evaluate_expr(expr, ns):
    """
    Try to evaluate the expression in the given namespace

    Returns either (value, True) if successful, or (expr, False) otherwise.

    Examples
    --------
    >>> assumptions = {'exp': DEFAULT_FUNCTIONS['exp'].pyfunc,
    ...                'inf': DEFAULT_CONSTANTS['inf'].value}
    >>> evaluate_expr('1/2', assumptions)
    (0.5, True)
    >>> evaluate_expr('exp(-inf)', assumptions)
    (0.0, True)
    >>> evaluate_expr('sin(2*pi*freq*t)', assumptions)
    ('sin(2*pi*freq*t)', False)
    >>> evaluate_expr('1/0', assumptions)
    ('1/0', False)
    """
    try:
        val = eval(expr, ns)
        return val, True
    except (NameError, ArithmeticError):
        return expr, False


def expression_complexity(expr, variables):
    return brian_ast(expr, variables).complexity


def optimise_statements(scalar_statements, vector_statements, variables, blockname=""):
    """
    Optimise a sequence of scalar and vector statements

    Performs the following optimisations:

    1. Constant evaluations (e.g. exp(0) to 1). See `evaluate_expr`.
    2. Arithmetic simplifications (e.g. 0*x to 0). See `ArithmeticSimplifier`, `collect`.
    3. Pulling out loop invariants (e.g. v*exp(-dt/tau) to a=exp(-dt/tau) outside the loop and v*a inside).
       See `Simplifier`.
    4. Boolean simplifications (allowing the replacement of expressions with booleans with a sequence of if/thens).
       See `Simplifier`.

    Parameters
    ----------
    scalar_statements : sequence of Statement
        Statements that only involve scalar values and should be evaluated in the scalar block.
    vector_statements : sequence of Statement
        Statements that involve vector values and should be evaluated in the vector block.
    variables : dict of (str, Variable)
        Definition of the types of the variables.
    blockname : str, optional
        Name of the block (used for LIO constant prefixes to avoid name clashes)

    Returns
    -------
    new_scalar_statements : sequence of Statement
        As above but with loop invariants pulled out from vector statements
    new_vector_statements : sequence of Statement
        Simplified/optimised versions of statements
    """
    boolvars = {
        k: v
        for k, v in variables.items()
        if hasattr(v, "dtype") and brian_dtype_from_dtype(v.dtype) == "boolean"
    }
    # We use the Simplifier class by rendering each expression, which generates new scalar statements
    # stored in the Simplifier object, and these are then added to the scalar statements.
    simplifier = Simplifier(variables, scalar_statements, extra_lio_prefix=blockname)
    new_vector_statements = []
    for stmt in vector_statements:
        # Carry out constant evaluation, arithmetic simplification and loop invariants
        new_expr = simplifier.render_expr(stmt.expr)
        new_stmt = Statement(
            stmt.var,
            stmt.op,
            new_expr,
            stmt.comment,
            dtype=stmt.dtype,
            constant=stmt.constant,
            subexpression=stmt.subexpression,
            scalar=stmt.scalar,
        )
        # Now check if boolean simplification can be carried out
        complexity_std = expression_complexity(new_expr, simplifier.variables)
        idents = get_identifiers(new_expr)
        used_boolvars = [var for var in boolvars if var in idents]
        if len(used_boolvars):
            # We want to iterate over all the possible assignments of boolean variables to values in (True, False)
            bool_space = [[False, True] for _ in used_boolvars]
            expanded_expressions = {}
            complexities = {}
            for bool_vals in itertools.product(*bool_space):
                # substitute those values into the expr and simplify (including potentially pulling out new
                # loop invariants)
                subs = {var: str(val) for var, val in zip(used_boolvars, bool_vals)}
                curexpr = word_substitute(new_expr, subs)
                curexpr = simplifier.render_expr(curexpr)
                key = tuple((var, val) for var, val in zip(used_boolvars, bool_vals))
                expanded_expressions[key] = curexpr
                complexities[key] = expression_complexity(curexpr, simplifier.variables)
            # See Statement for details on these
            new_stmt.used_boolean_variables = used_boolvars
            new_stmt.boolean_simplified_expressions = expanded_expressions
            new_stmt.complexity_std = complexity_std
            new_stmt.complexities = complexities
        new_vector_statements.append(new_stmt)
    # Generate additional scalar statements for the loop invariants
    new_scalar_statements = copy.copy(scalar_statements)
    for expr, name in simplifier.loop_invariants.items():
        dtype_name = simplifier.loop_invariant_dtypes[name]
        if dtype_name == "boolean":
            dtype = bool
        elif dtype_name == "integer":
            dtype = int
        else:
            dtype = prefs.core.default_float_dtype
        new_stmt = Statement(
            name,
            ":=",
            expr,
            "",
            dtype=dtype,
            constant=True,
            subexpression=False,
            scalar=True,
        )
        new_scalar_statements.append(new_stmt)
    return new_scalar_statements, new_vector_statements


def _replace_with_zero(zero_node, node):
    """
    Helper function to return a "zero node" of the correct type.

    Parameters
    ----------
    zero_node : `ast.Constant`
        The node to replace
    node : `ast.Node`
        The node that determines the type

    Returns
    -------
    zero_node : `ast.Constant`
        The original ``zero_node`` with its value replaced by 0 or 0.0.
    """
    # must not change the dtype of the output,
    # e.g. handle 0/float->0.0 and 0.0/int->0.0
    zero_node.dtype = node.dtype
    if node.dtype == "integer":
        zero_node.value = 0
    else:
        zero_node.value = prefs.core.default_float_dtype(0.0)
    return zero_node


class ArithmeticSimplifier(BrianASTRenderer):
    """
    Carries out the following arithmetic simplifications:

    1. Constant evaluation (e.g. exp(0)=1) by attempting to evaluate the expression in an "assumptions namespace"
    2. Binary operators, e.g. 0*x=0, 1*x=x, etc. You have to take care that the dtypes match here, e.g.
       if x is an integer, then 1.0*x shouldn't be replaced with x but left as 1.0*x.

    Parameters
    ----------
    variables : dict of (str, Variable)
        Usual definition of variables.
    assumptions : sequence of str
        Additional assumptions that can be used in simplification, each assumption is a string statement.
        These might be the scalar statements for example.
    """

    def __init__(self, variables):
        BrianASTRenderer.__init__(self, variables, copy_variables=False)
        self.assumptions = []
        self.assumptions_ns = dict(defaults_ns)
        self.bast_renderer = BrianASTRenderer(variables, copy_variables=False)

    def render_node(self, node):
        """
        Assumes that the node has already been fully processed by BrianASTRenderer
        """
        if not hasattr(node, "simplified"):
            node = super().render_node(node)
            node.simplified = True
        # can't evaluate vector expressions, so abandon in this case
        if not node.scalar:
            return node
        # No evaluation necessary for simple names or numbers
        if node.__class__.__name__ in ["Name", "NameConstant", "Num", "Constant"]:
            return node
        # Don't evaluate stateful nodes (e.g. those containing a rand() call)
        if not node.stateless:
            return node
        # try fully evaluating using assumptions
        expr = NodeRenderer().render_node(node)
        val, evaluated = evaluate_expr(expr, self.assumptions_ns)
        if evaluated:
            if node.dtype == "boolean":
                val = bool(val)
                if hasattr(ast, "Constant"):
                    newnode = ast.Constant(val)
                elif hasattr(ast, "NameConstant"):
                    newnode = ast.NameConstant(val)
                else:
                    # None is the expression context, we don't use it so we just set to None
                    newnode = ast.Name(repr(val), None)
            elif node.dtype == "integer":
                val = int(val)
            else:
                val = prefs.core.default_float_dtype(val)
            if node.dtype != "boolean":
                newnode = ast.Constant(val)
            newnode.dtype = node.dtype
            newnode.scalar = True
            newnode.stateless = node.stateless
            newnode.complexity = 0
            return newnode
        return node

    def render_BinOp(self, node):
        if node.dtype == "float":  # only try to collect float type nodes
            if node.op.__class__.__name__ in [
                "Mult",
                "Div",
                "Add",
                "Sub",
            ] and not hasattr(node, "collected"):
                newnode = self.bast_renderer.render_node(collect(node))
                newnode.collected = True
                return self.render_node(newnode)
        left = node.left = self.render_node(node.left)
        right = node.right = self.render_node(node.right)
        node = super().render_BinOp(node)
        op = node.op
        # Handle multiplication by 0 or 1
        if op.__class__.__name__ == "Mult":
            for operand, other in [(left, right), (right, left)]:
                if operand.__class__.__name__ in ["Num", "Constant"]:
                    op_value = operand.value
                    if op_value == 0:
                        # Do not remove stateful functions
                        if node.stateless:
                            return _replace_with_zero(operand, node)
                    if op_value == 1:
                        # only simplify this if the type wouldn't be cast by the operation
                        if (
                            dtype_hierarchy[operand.dtype]
                            <= dtype_hierarchy[other.dtype]
                        ):
                            return other
        # Handle division by 1, or 0/x
        elif op.__class__.__name__ == "Div":
            if (
                left.__class__.__name__ in ["Num", "Constant"] and left.value == 0
            ):  # 0/x
                if node.stateless:
                    # Do not remove stateful functions
                    return _replace_with_zero(left, node)
            if (
                right.__class__.__name__ in ["Num", "Constant"] and right.value == 1
            ):  # x/1
                # only simplify this if the type wouldn't be cast by the operation
                if dtype_hierarchy[right.dtype] <= dtype_hierarchy[left.dtype]:
                    return left
        elif op.__class__.__name__ == "FloorDiv":
            if (
                left.__class__.__name__ in ["Num", "Constant"] and left.value == 0
            ):  # 0//x
                if node.stateless:
                    # Do not remove stateful functions
                    return _replace_with_zero(left, node)
            # Only optimise floor division by 1 if both numbers are integers,
            # for floating point values, floor division by 1 changes the value,
            # and division by 1.0 can change the type for an integer value
            if (
                left.dtype == right.dtype == "integer"
                and right.__class__.__name__ in ["Num", "Constant"]
                and right.value == 1
            ):  # x//1
                return left
        # Handle addition of 0
        elif op.__class__.__name__ == "Add":
            for operand, other in [(left, right), (right, left)]:
                if (
                    operand.__class__.__name__ in ["Num", "Constant"]
                    and operand.value == 0
                ):
                    # only simplify this if the type wouldn't be cast by the operation
                    if dtype_hierarchy[operand.dtype] <= dtype_hierarchy[other.dtype]:
                        return other
        # Handle subtraction of 0
        elif op.__class__.__name__ == "Sub":
            if right.__class__.__name__ in ["Num", "Constant"] and right.value == 0:
                # only simplify this if the type wouldn't be cast by the operation
                if dtype_hierarchy[right.dtype] <= dtype_hierarchy[left.dtype]:
                    return left

        # simplify e.g. 2*float to 2.0*float to make things more explicit: not strictly necessary
        # but might be useful for some codegen targets
        if node.dtype == "float" and op.__class__.__name__ in [
            "Mult",
            "Add",
            "Sub",
            "Div",
        ]:
            for subnode in [node.left, node.right]:
                if subnode.__class__.__name__ in ["Num", "Constant"] and not (
                    subnode.value is True or subnode.value is False
                ):
                    subnode.dtype = "float"
                    subnode.value = prefs.core.default_float_dtype(subnode.value)
        return node


class Simplifier(BrianASTRenderer):
    """
    Carry out arithmetic simplifications (see `ArithmeticSimplifier`) and loop invariants

    Parameters
    ----------
    variables : dict of (str, Variable)
        Usual definition of variables.
    scalar_statements : sequence of Statement
        Predefined scalar statements that can be used as part of simplification

    Notes
    -----

    After calling `render_expr` on a sequence of expressions (coming from vector statements typically),
    this object will have some new attributes:

    ``loop_invariants`` : OrderedDict of (expression, varname)
        varname will be of the form ``_lio_N`` where ``N`` is some integer, and the expressions will be
        strings that correspond to scalar-only expressions that can be evaluated outside of the vector
        block.
    ``loop_invariant_dtypes`` : dict of (varname, dtypename)
        dtypename will be one of ``'boolean'``, ``'integer'``, ``'float'``.
    """

    def __init__(self, variables, scalar_statements, extra_lio_prefix=""):
        BrianASTRenderer.__init__(self, variables, copy_variables=False)
        self.loop_invariants = OrderedDict()
        self.loop_invariant_dtypes = {}
        self.value = 0
        self.node_renderer = NodeRenderer()
        self.arithmetic_simplifier = ArithmeticSimplifier(variables)
        self.scalar_statements = scalar_statements
        if extra_lio_prefix is None:
            extra_lio_prefix = ""
        if len(extra_lio_prefix):
            extra_lio_prefix = f"{extra_lio_prefix}_"
        self.extra_lio_prefix = extra_lio_prefix

    def render_expr(self, expr):
        node = brian_ast(expr, self.variables)
        node = self.arithmetic_simplifier.render_node(node)
        node = self.render_node(node)
        return self.node_renderer.render_node(node)

    def render_node(self, node):
        """
        Assumes that the node has already been fully processed by BrianASTRenderer
        """
        # can we pull this out?
        if node.scalar and node.complexity > 0:
            expr = self.node_renderer.render_node(
                self.arithmetic_simplifier.render_node(node)
            )
            if expr in self.loop_invariants:
                name = self.loop_invariants[expr]
            else:
                self.value += 1
                name = f"_lio_{self.extra_lio_prefix}{str(self.value)}"
                self.loop_invariants[expr] = name
                self.loop_invariant_dtypes[name] = node.dtype
                numpy_dtype = {
                    "boolean": bool,
                    "integer": int,
                    "float": prefs.core.default_float_dtype,
                }[node.dtype]
                self.variables[name] = AuxiliaryVariable(
                    name, dtype=numpy_dtype, scalar=True
                )
            # None is the expression context, we don't use it so we just set to None
            newnode = ast.Name(name, None)
            newnode.scalar = True
            newnode.dtype = node.dtype
            newnode.complexity = 0
            newnode.stateless = node.stateless
            return newnode
        # otherwise, render node as usual
        return super().render_node(node)


def reduced_node(terms, op):
    """
    Reduce a sequence of terms with the given operator

    For examples, if terms were [a, b, c] and op was multiplication then the reduction would be (a*b)*c.

    Parameters
    ----------
    terms : list
        AST nodes.
    op : AST node
        Could be `ast.Mult` or `ast.Add`.

    Examples
    --------
    >>> import ast
    >>> nodes = [ast.Name(id='x'), ast.Name(id='y'), ast.Name(id='z')]
    >>> ast.unparse(reduced_node(nodes, ast.Mult))
    'x * y * z'
    >>> nodes = [ast.Name(id='x')]
    >>> ast.unparse(reduced_node(nodes, ast.Add))
    'x'
    """
    # Remove None terms
    terms = [term for term in terms if term is not None]
    if not len(terms):
        return None
    return reduce(lambda left, right: ast.BinOp(left, op(), right), terms)


def cancel_identical_terms(primary, inverted):
    """
    Cancel terms in a collection, e.g. a+b-a should be cancelled to b

    Simply renders the nodes into expressions and removes whenever there is a common expression
    in primary and inverted.

    Parameters
    ----------
    primary : list of AST nodes
        These are the nodes that are positive with respect to the operator, e.g.
        in x*y/z it would be [x, y].
    inverted : list of AST nodes
        These are the nodes that are inverted with respect to the operator, e.g.
        in x*y/z it would be [z].

    Returns
    -------
    primary : list of AST nodes
        Primary nodes after cancellation
    inverted : list of AST nodes
        Inverted nodes after cancellation
    """
    nr = NodeRenderer()
    expressions = {node: nr.render_node(node) for node in primary}
    expressions.update({node: nr.render_node(node) for node in inverted})
    new_primary = []
    inverted_expressions = [expressions[term] for term in inverted]
    for term in primary:
        expr = expressions[term]
        if expr in inverted_expressions and term.stateless:
            new_inverted = []
            for iterm in inverted:
                if expressions[iterm] == expr:
                    expr = ""  # handled
                else:
                    new_inverted.append(iterm)
            inverted = new_inverted
            inverted_expressions = [expressions[term] for term in inverted]
        else:
            new_primary.append(term)
    return new_primary, inverted


def collect(node):
    """
    Attempts to collect commutative operations into one and simplifies them.

    For example, if x and y are scalars, and z is a vector, then (x*z)*y should
    be rewritten as (x*y)*z to minimise the number of vector operations. Similarly,
    ((x*2)*3)*4 should be rewritten as x*24.

    Works for either multiplication/division or addition/subtraction nodes.

    The final output is a subexpression of the following maximal form:

        (((numerical_value*(product of scalars))/(product of scalars))*(product of vectors))/(product of vectors)

    Any possible cancellations will have been done.

    Parameters
    ----------
    node : Brian AST node
        The node to be collected/simplified.

    Returns
    -------
    node : Brian AST node
        Simplified node.
    """
    node.collected = True
    orignode_dtype = node.dtype
    # we only work on */ or +- ops, which are both BinOp
    if node.__class__.__name__ != "BinOp":
        return node
    # primary would be the * or + nodes, and inverted would be the / or - nodes
    terms_primary = []
    terms_inverted = []
    # we handle both multiplicative and additive nodes in the same way by using these variables
    if node.op.__class__.__name__ in ["Mult", "Div"]:
        op_primary = ast.Mult
        op_inverted = ast.Div
        op_null = prefs.core.default_float_dtype(1.0)  # the identity for the operator
        op_py_primary = lambda x, y: x * y
        op_py_inverted = lambda x, y: x / y
    elif node.op.__class__.__name__ in ["Add", "Sub"]:
        op_primary = ast.Add
        op_inverted = ast.Sub
        op_null = prefs.core.default_float_dtype(0.0)
        op_py_primary = lambda x, y: x + y
        op_py_inverted = lambda x, y: x - y
    else:
        return node
    if node.dtype == "integer":
        op_null_with_dtype = int(op_null)
    else:
        op_null_with_dtype = op_null
    # recursively collect terms into the terms_primary and terms_inverted lists
    collect_commutative(node, op_primary, op_inverted, terms_primary, terms_inverted)
    x = op_null
    # extract the numerical nodes and fully evaluate
    remaining_terms_primary = []
    remaining_terms_inverted = []
    for term in terms_primary:
        if term.__class__.__name__ == "Num":
            x = op_py_primary(x, term.n)
        elif term.__class__.__name__ == "Constant":
            x = op_py_primary(x, term.value)
        else:
            remaining_terms_primary.append(term)
    for term in terms_inverted:
        if term.__class__.__name__ == "Num":
            x = op_py_inverted(x, term.n)
        elif term.__class__.__name__ == "Constant":
            x = op_py_inverted(x, term.value)
        else:
            remaining_terms_inverted.append(term)
    # if the fully evaluated node is just the identity/null element then we
    # don't have to make it into an explicit term
    if x != op_null:
        num_node = ast.Constant(x)
    else:
        num_node = None
    terms_primary = remaining_terms_primary
    terms_inverted = remaining_terms_inverted
    node = num_node
    for scalar in (True, False):
        primary_terms = [term for term in terms_primary if term.scalar == scalar]
        inverted_terms = [term for term in terms_inverted if term.scalar == scalar]
        primary_terms, inverted_terms = cancel_identical_terms(
            primary_terms, inverted_terms
        )

        # produce nodes that are the reduction of the operator on these subsets
        prod_primary = reduced_node(primary_terms, op_primary)
        prod_inverted = reduced_node(inverted_terms, op_primary)

        # construct the simplest version of the fully simplified node (only doing operations where necessary)
        node = reduced_node([node, prod_primary], op_primary)
        if prod_inverted is not None:
            if node is None:
                node = ast.Constant(op_null_with_dtype)
            node = ast.BinOp(node, op_inverted(), prod_inverted)

    if node is None:  # everything cancelled
        node = ast.Constant(op_null_with_dtype)
    if (
        hasattr(node, "dtype")
        and dtype_hierarchy[node.dtype] < dtype_hierarchy[orignode_dtype]
    ):
        node = ast.BinOp(ast.Constant(op_null_with_dtype), op_primary(), node)
    node.collected = True
    return node


def collect_commutative(
    node, primary, inverted, terms_primary, terms_inverted, add_to_inverted=False
):
    # This function is called recursively, so we use add_to_inverted to keep track of whether or not
    # we're working in the numerator/denominator (for multiplicative nodes, equivalent for additive).
    op_primary = node.op.__class__ is primary
    # this should only be called with node a BinOp of type primary or inverted
    # left_exact is the condition that we can collect terms (we can do it with floats or add/sub,
    # but not integer mult/div - the reason being that for C-style division e.g. 3/(4/3)!=(3*3)/4
    left_exact = node.left.dtype == "float" or (
        hasattr(node.left, "op") and node.left.op.__class__.__name__ in ["Add", "Sub"]
    )
    if (
        node.left.__class__.__name__ == "BinOp"
        and node.left.op.__class__ in [primary, inverted]
        and left_exact
    ):
        collect_commutative(
            node.left,
            primary,
            inverted,
            terms_primary,
            terms_inverted,
            add_to_inverted=add_to_inverted,
        )
    else:
        if add_to_inverted:
            terms_inverted.append(node.left)
        else:
            terms_primary.append(node.left)
    right_exact = node.right.dtype == "float" or (
        hasattr(node.right, "op") and node.right.op.__class__.__name__ in ["Add", "Sub"]
    )
    if (
        node.right.__class__.__name__ == "BinOp"
        and node.right.op.__class__ in [primary, inverted]
        and right_exact
    ):
        if node.op.__class__ is primary:
            collect_commutative(
                node.right,
                primary,
                inverted,
                terms_primary,
                terms_inverted,
                add_to_inverted=add_to_inverted,
            )
        else:
            collect_commutative(
                node.right,
                primary,
                inverted,
                terms_primary,
                terms_inverted,
                add_to_inverted=not add_to_inverted,
            )
    else:
        if (not add_to_inverted and op_primary) or (add_to_inverted and not op_primary):
            terms_primary.append(node.right)
        else:
            terms_inverted.append(node.right)



import logging
import re

import pytest
from numpy.testing import assert_equal

from brian2 import *
from brian2.core.variables import ArrayVariable, Constant, Variable
from brian2.stateupdaters.base import UnsupportedEquationsException
from brian2.tests.utils import assert_allclose, exc_isinstance
from brian2.utils.logger import catch_logs


@pytest.mark.codegen_independent
def test_explicit_stateupdater_parsing():
    """
    Test the parsing of explicit state updater descriptions.
    """
    # These are valid descriptions and should not raise errors
    updater = ExplicitStateUpdater("x_new = x + dt * f(x, t)")
    updater(Equations("dv/dt = -v / tau : 1"))
    updater = ExplicitStateUpdater(
        """
        x2 = x + dt * f(x, t)
        x_new = x2
        """
    )
    updater(Equations("dv/dt = -v / tau : 1"))
    updater = ExplicitStateUpdater(
        """
        x1 = g(x, t) * dW
        x2 = x + dt * f(x, t)
        x_new = x1 + x2
        """,
        stochastic="multiplicative",
    )
    updater(Equations("dv/dt = -v / tau + v * xi * tau**-.5: 1"))

    updater = ExplicitStateUpdater(
        """
        x_support = x + dt*f(x, t) + dt**.5 * g(x, t)
        g_support = g(x_support, t)
        k = 1/(2*dt**.5)*(g_support - g(x, t))*(dW**2)
        x_new = x + dt*f(x,t) + g(x, t) * dW + k
        """,
        stochastic="multiplicative",
    )
    updater(Equations("dv/dt = -v / tau + v * xi * tau**-.5: 1"))

    # Examples of failed parsing
    # No x_new = ... statement
    with pytest.raises(SyntaxError):
        ExplicitStateUpdater("x = x + dt * f(x, t)")
    # Not an assigment
    with pytest.raises(SyntaxError):
        ExplicitStateUpdater(
            """
            2 * x
            x_new = x + dt * f(x, t)
            """
        )

    # doesn't separate into stochastic and non-stochastic part
    updater = ExplicitStateUpdater("x_new = x + dt * f(x, t) * g(x, t) * dW")
    with pytest.raises(ValueError):
        updater(Equations(""))


@pytest.mark.codegen_independent
def test_non_autonomous_equations():
    # Check that non-autonmous equations are handled correctly in multi-step
    # updates
    updater = ExplicitStateUpdater("x_new = f(x, t + 0.5*dt)")
    update_step = updater(Equations("dv/dt = t : 1"))  # Not a valid equation but...
    # very crude test
    assert "0.5*dt" in update_step


@pytest.mark.codegen_independent
def test_str_repr():
    """
    Assure that __str__ and __repr__ do not raise errors
    """
    for integrator in [linear, euler, rk2, rk4]:
        assert len(str(integrator))
        assert len(repr(integrator))


@pytest.mark.codegen_independent
def test_multiple_noise_variables_basic():
    # Very basic test, only to make sure that stochastic state updaters handle
    # multiple noise variables at all
    eqs = Equations(
        """
        dv/dt = -v / (10*ms) + xi_1 * ms ** -.5 : 1
        dw/dt = -w / (10*ms) + xi_2 * ms ** -.5 : 1
        """
    )
    for method in [euler, heun, milstein]:
        code = method(eqs, {})
        assert "xi_1" in code
        assert "xi_2" in code


def test_multiple_noise_variables_extended():
    # Some actual simulations with multiple noise variables
    eqs = """
          dx/dt = y : 1
          dy/dt = - 1*ms**-1*y - 40*ms**-2*x : Hz
    """
    all_eqs_noise = [
        """
        dx/dt = y : 1
        dy/dt = noise_factor*ms**-1.5*xi_1 + noise_factor*ms**-1.5*xi_2
                - 1*ms**-1*y - 40*ms**-2*x : Hz
        """,
        """
        dx/dt = y + noise_factor*ms**-0.5*xi_1: 1
        dy/dt = noise_factor*ms**-1.5*xi_2
                - 1*ms**-1*y - 40*ms**-2*x : Hz
        """,
    ]
    G = NeuronGroup(2, eqs, method="euler")
    G.x = [0.5, 1]
    G.y = [0, 0.5] * Hz
    mon = StateMonitor(G, ["x", "y"], record=True)
    net = Network(G, mon)
    net.run(10 * ms)
    no_noise_x, no_noise_y = mon.x[:], mon.y[:]

    for eqs_noise in all_eqs_noise:
        for method_name, method in [("euler", euler), ("heun", heun)]:
            with catch_logs("WARNING"):
                G = NeuronGroup(2, eqs_noise, method=method)
                G.x = [0.5, 1]
                G.y = [0, 0.5] * Hz
                mon = StateMonitor(G, ["x", "y"], record=True)
                net = Network(G, mon)
                # We run it deterministically, but still we'd detect major errors (e.g.
                # non-stochastic terms that are added twice, see #330
                net.run(10 * ms, namespace={"noise_factor": 0.0})
            assert_allclose(
                mon.x[:],
                no_noise_x,
                err_msg=f"Method {method_name} gave incorrect results",
            )
            assert_allclose(
                mon.y[:],
                no_noise_y,
                err_msg=f"Method {method_name} gave incorrect results",
            )


def test_multiple_noise_variables_deterministic_noise(fake_randn_randn_fixture):
    all_eqs = [
        """
        dx/dt = y : 1
        dy/dt = -y / (10*ms) + dt**-.5*0.5*ms**-1.5 + dt**-.5*0.5*ms**-1.5: Hz
        """,
        """
        dx/dt = y + dt**-.5*0.5*ms**-0.5: 1
        dy/dt = -y / (10*ms) + dt**-.5*0.5 * ms**-1.5 : Hz
        """,
    ]
    all_eqs_noise = [
        """
        dx/dt = y : 1
        dy/dt = -y / (10*ms) + xi_1 * ms**-1.5 + xi_2 * ms**-1.5: Hz
        """,
        """
        dx/dt = y + xi_1*ms**-0.5: 1
        dy/dt = -y / (10*ms) + xi_2 * ms**-1.5 : Hz
        """,
    ]
    for eqs, eqs_noise in zip(all_eqs, all_eqs_noise):
        G = NeuronGroup(2, eqs, method="euler")
        G.x = [5, 17]
        G.y = [25, 5] * Hz
        mon = StateMonitor(G, ["x", "y"], record=True)
        net = Network(G, mon)
        net.run(10 * ms)
        no_noise_x, no_noise_y = mon.x[:], mon.y[:]

        for method_name, method in [("euler", euler), ("heun", heun)]:
            with catch_logs("WARNING"):
                G = NeuronGroup(2, eqs_noise, method=method)
                G.x = [5, 17]
                G.y = [25, 5] * Hz
                mon = StateMonitor(G, ["x", "y"], record=True)
                net = Network(G, mon)
                net.run(10 * ms)
            assert_allclose(
                mon.x[:],
                no_noise_x,
                err_msg=f"Method {method_name} gave incorrect results",
            )
            assert_allclose(
                mon.y[:],
                no_noise_y,
                err_msg=f"Method {method_name} gave incorrect results",
            )


@pytest.mark.codegen_independent
def test_multiplicative_noise():
    # Noise is not multiplicative (constant over time step)
    ta = TimedArray([0, 1], dt=defaultclock.dt * 10)
    Eq = Equations("dv/dt = ta(t)*xi*(5*ms)**-0.5 :1")
    group = NeuronGroup(1, Eq, method="euler")
    net = Network(group)
    net.run(0 * ms)  # no error

    # Noise is multiplicative (multiplied with time-varying variable)
    Eq1 = Equations("dv/dt = v*xi*(5*ms)**-0.5 :1")
    group1 = NeuronGroup(1, Eq1, method="euler")
    net1 = Network(group1)
    with pytest.raises(BrianObjectException) as exc:
        net1.run(0 * ms)
    assert exc_isinstance(exc, UnsupportedEquationsException)

    # Noise is multiplicative (multiplied with time)
    Eq2 = Equations("dv/dt = (t/ms)*xi*(5*ms)**-0.5 :1")
    group2 = NeuronGroup(1, Eq2, method="euler")
    net2 = Network(group2)
    with pytest.raises(BrianObjectException) as exc:
        net2.run(0 * ms)
    assert exc_isinstance(exc, UnsupportedEquationsException)

    # Noise is multiplicative (multiplied with time-varying variable)
    Eq3 = Equations(
        """
        dv/dt = w*xi*(5*ms)**-0.5 :1
        dw/dt = -w/(10*ms) : 1
        """
    )
    group3 = NeuronGroup(1, Eq3, method="euler")
    net3 = Network(group3)
    with pytest.raises(BrianObjectException) as exc:
        net3.run(0 * ms)
    assert exc_isinstance(exc, UnsupportedEquationsException)

    # One of the equations has multiplicative noise
    Eq4 = Equations(
        """
        dv/dt = xi_1*(5*ms)**-0.5 : 1
        dw/dt = (t/ms)*xi_2*(5*ms)**-0.5 :1
        """
    )
    group4 = NeuronGroup(1, Eq4, method="euler")
    net4 = Network(group4)
    with pytest.raises(BrianObjectException) as exc:
        net4.run(0 * ms)
    assert exc_isinstance(exc, UnsupportedEquationsException)

    # One of the equations has multiplicative noise
    Eq5 = Equations(
        """
        dv/dt = xi_1*(5*ms)**-0.5 : 1
        dw/dt = v*xi_2*(5*ms)**-0.5 :1
        """
    )
    group5 = NeuronGroup(1, Eq5, method="euler")
    net5 = Network(group4)
    with pytest.raises(BrianObjectException) as exc:
        net5.run(0 * ms)
    assert exc_isinstance(exc, UnsupportedEquationsException)


def test_pure_noise_deterministic(fake_randn_randn_fixture):
    sigma = 3.0
    eqs = Equations("dx/dt = sigma*xi/sqrt(ms) : 1")
    dt = 0.1 * ms
    for method in ["euler", "heun", "milstein"]:
        G = NeuronGroup(1, eqs, dt=dt, method=method)
        run(10 * dt)
        assert_allclose(
            G.x,
            sqrt(dt) * sigma * 0.5 / sqrt(1 * ms) * 10,
            err_msg=f"method {method} did not give the expected result",
        )


@pytest.mark.codegen_independent
def test_temporary_variables():
    """
    Make sure that the code does the distinction between temporary variables
    in the state updater description and external variables used in the
    equations.
    """
    # Use a variable name that is used in the state updater description
    k_2 = 5
    eqs = Equations("dv/dt = -(v + k_2)/(10*ms) : 1")
    converted = rk4(eqs)

    # Use a non-problematic name
    k_var = 5
    eqs = Equations("dv/dt = -(v + k_var)/(10*ms) : 1")
    converted2 = rk4(eqs)

    # Make sure that the two formulations result in the same code
    assert converted == converted2.replace("k_var", "k_2")


@pytest.mark.codegen_independent
def test_temporary_variables2():
    """
    Make sure that the code does the distinction between temporary variables
    in the state updater description and external variables used in the
    equations.
    """
    tau = 10 * ms
    # Use a variable name that is used in the state updater description
    k = 5
    eqs = Equations("dv/dt = -v/tau + k*xi*tau**-0.5: 1")
    converted = milstein(eqs)

    # Use a non-problematic name
    k_var = 5
    eqs = Equations("dv/dt = -v/tau + k_var*xi*tau**-0.5: 1")
    converted2 = milstein(eqs)

    # Make sure that the two formulations result in the same code
    assert converted == converted2.replace("k_var", "k")


@pytest.mark.codegen_independent
def test_integrator_code():
    """
    Check whether the returned abstract code is as expected.
    """
    # A very simple example where the abstract code should always look the same
    eqs = Equations("dv/dt = -v / (1 * second) : 1")

    # Only test very basic stuff (expected number of lines and last line)
    for integrator, lines in zip([linear, euler, rk2, rk4], [2, 2, 3, 6]):
        code_lines = integrator(eqs).split("\n")
        err_msg = (
            f"Returned code for integrator {integrator.__class__.__name__} had"
            f" {len(code_lines)} lines instead of {int(lines)}"
        )
        assert len(code_lines) == lines, err_msg
        assert code_lines[-1] == "v = _v"

    # Make sure that it isn't a problem to use 'x', 'f' and 'g'  as variable
    # names, even though they are also used in state updater descriptions.
    # The resulting code should be identical when replacing x by x0 (and ..._x by
    # ..._x0)
    for varname in ["x", "f", "g"]:
        # We use a very similar names here to avoid slightly re-arranged
        # expressions due to alphabetical sorting of terms in
        # multiplications, etc.
        eqs_v = Equations(f"d{varname}0/dt = -{varname}0 / (1 * second) : 1")
        eqs_var = Equations(f"d{varname}/dt = -{varname} / (1 * second) : 1")
        for integrator in [linear, euler, rk2, rk4]:
            code_v = integrator(eqs_v)
            code_var = integrator(eqs_var)
            # Re-substitute the variable names in the output
            code_var = re.sub(rf"\b{varname}\b", f"{varname}0", code_var)
            code_var = re.sub(rf"\b(\w*)_{varname}\b", rf"\1_{varname}0", code_var)
            assert code_var == code_v, f"'{code_var}' does not match '{code_v}'"


@pytest.mark.codegen_independent
def test_integrator_code2():
    """
    Test integration for a simple model with several state variables.
    """
    eqs = Equations(
        """
        dv/dt=(ge+gi-v)/tau : volt
        dge/dt=-ge/taue : volt
        dgi/dt=-gi/taui : volt
        """
    )
    euler_integration = euler(eqs)
    lines = sorted(euler_integration.split("\n"))
    # Do a very basic check that the right variables are used in every line
    for varname, line in zip(["_ge", "_gi", "_v", "ge", "gi", "v"], lines):
        assert line.startswith(
            f"{varname} = "
        ), f'line "{line}" does not start with {varname}'
    for variables, line in zip(
        [
            ["dt", "ge", "taue"],
            ["dt", "gi", "taui"],
            ["dt", "ge", "gi", "v", "tau"],
            ["_ge"],
            ["_gi"],
            ["_v"],
        ],
        lines,
    ):
        rhs = line.split("=")[1]
        for variable in variables:
            assert variable in rhs, f'{variable} not in RHS: "{rhs}"'


@pytest.mark.codegen_independent
def test_illegal_calls():
    eqs = Equations("dv/dt = -v / (10*ms) : 1")
    clock = Clock(dt=0.1 * ms)
    variables = {
        "v": ArrayVariable(
            name="name",
            size=10,
            owner=None,
            device=None,
            dtype=np.float64,
            constant=False,
        ),
        "t": clock.variables["t"],
        "dt": clock.variables["dt"],
    }
    with pytest.raises(TypeError):
        StateUpdateMethod.apply_stateupdater(eqs, variables, object())
    with pytest.raises(TypeError):
        StateUpdateMethod.apply_stateupdater(
            eqs, variables, group_name="my_name", method=object()
        )
    with pytest.raises(TypeError):
        StateUpdateMethod.apply_stateupdater(eqs, variables, [object(), "euler"])
    with pytest.raises(TypeError):
        StateUpdateMethod.apply_stateupdater(
            eqs, variables, group_name="my_name", method=[object(), "euler"]
        )


def check_integration(eqs, variables, can_integrate):
    # can_integrate maps integrators to True/False/None
    # True/False means that the integrator should/should not integrate the equations
    # None means that it *might* integrate the equations (only needed for the
    # exact integration, since it can depend on the sympy version)
    for integrator, able in can_integrate.items():
        try:
            integrator(eqs, variables)
            if able is False:
                raise AssertionError(
                    "Should not be able to integrate these "
                    f"equations (equations: '{eqs}') with "
                    f"integrator {integrator.__class__.__name__}"
                )
        except UnsupportedEquationsException:
            if able is True:
                raise AssertionError(
                    "Should be able to integrate these "
                    f"equations (equations: '{eqs}') with "
                    f"integrator {integrator.__class__.__name__}"
                )


@pytest.mark.codegen_independent
def test_priority():
    updater = ExplicitStateUpdater("x_new = x + dt * f(x, t)")
    # Equations that work for the state updater
    eqs = Equations("dv/dt = -v / (10*ms) : 1")
    clock = Clock(dt=0.1 * ms)
    variables = {
        "v": ArrayVariable(
            name="name",
            size=10,
            owner=None,
            device=None,
            dtype=np.float64,
            constant=False,
        ),
        "w": ArrayVariable(
            name="name",
            size=10,
            owner=None,
            device=None,
            dtype=np.float64,
            constant=False,
        ),
        "t": clock.variables["t"],
        "dt": clock.variables["dt"],
    }
    updater(eqs, variables)  # should not raise an error

    # External parameter in the coefficient, linear integration should work
    param = 1
    eqs = Equations("dv/dt = -param * v / (10*ms) : 1")
    updater(eqs, variables)  # should not raise an error
    can_integrate = {
        linear: True,
        euler: True,
        exponential_euler: True,
        rk2: True,
        rk4: True,
        heun: True,
        milstein: True,
    }

    check_integration(eqs, variables, can_integrate)

    # Constant equation, should work for all except linear (see #1010)
    param = 1
    eqs = Equations(
        """dv/dt = 10*Hz : 1
                       dw/dt = -v/(10*ms) : 1"""
    )
    updater(eqs, variables)  # should not raise an error
    can_integrate = {
        linear: None,
        euler: True,
        exponential_euler: True,
        rk2: True,
        rk4: True,
        heun: True,
        milstein: True,
    }

    check_integration(eqs, variables, can_integrate)

    # Equations resulting in complex linear solution for older versions of sympy
    eqs = Equations(
        """
        dv/dt      = (ge+gi-(v+49*mV))/(20*ms) : volt
        dge/dt     = -ge/(5*ms) : volt
        dgi/dt     = Dgi/(5*ms) : volt
        dDgi/dt    = ((-2./5) * Dgi - (1./5**2)*gi)/(10*ms) : volt
        """
    )
    can_integrate = {
        linear: None,
        euler: True,
        exponential_euler: True,
        rk2: True,
        rk4: True,
        heun: True,
        milstein: True,
    }
    check_integration(eqs, variables, can_integrate)

    # Equation with additive noise
    eqs = Equations("dv/dt = -v / (10*ms) + xi/(10*ms)**.5 : 1")
    with pytest.raises(UnsupportedEquationsException):
        updater(eqs, variables)

    can_integrate = {
        linear: False,
        euler: True,
        exponential_euler: False,
        rk2: False,
        rk4: False,
        heun: True,
        milstein: True,
    }

    check_integration(eqs, variables, can_integrate)

    # Equation with multiplicative noise
    eqs = Equations("dv/dt = -v / (10*ms) + v*xi/(10*ms)**.5 : 1")
    with pytest.raises(UnsupportedEquationsException):
        updater(eqs, variables)

    can_integrate = {
        linear: False,
        euler: False,
        exponential_euler: False,
        rk2: False,
        rk4: False,
        heun: True,
        milstein: True,
    }

    check_integration(eqs, variables, can_integrate)


@pytest.mark.codegen_independent
def test_registration():
    """
    Test state updater registration.
    """
    # Save state before tests
    before = dict(StateUpdateMethod.stateupdaters)

    lazy_updater = ExplicitStateUpdater("x_new = x")
    StateUpdateMethod.register("lazy", lazy_updater)

    # Trying to register again
    with pytest.raises(ValueError):
        StateUpdateMethod.register("lazy", lazy_updater)

    # Trying to register something that is not a state updater
    with pytest.raises(ValueError):
        StateUpdateMethod.register("foo", "just a string")

    # Trying to register with an invalid index
    with pytest.raises(TypeError):
        StateUpdateMethod.register("foo", lazy_updater, index="not an index")

    # reset to state before the test
    StateUpdateMethod.stateupdaters = before


@pytest.mark.codegen_independent
def test_determination():
    """
    Test the determination of suitable state updaters.
    """
    # To save some typing
    apply_stateupdater = StateUpdateMethod.apply_stateupdater

    eqs = Equations("dv/dt = -v / (10*ms) : 1")
    # Just make sure that state updaters know about the two state variables
    variables = {"v": Variable(name="v"), "w": Variable(name="w")}

    # all methods should work for these equations.
    # First, specify them explicitly (using the object)
    for integrator in (
        linear,
        euler,
        exponential_euler,  # TODO: Removed "independent" here due to the issue in sympy 0.7.4
        rk2,
        rk4,
        heun,
        milstein,
    ):
        with catch_logs() as logs:
            returned = apply_stateupdater(eqs, variables, method=integrator)
            assert (
                len(logs) == 0
            ), f"Got {len(logs)} unexpected warnings: {str([l[2] for l in logs])}"

    # Equation with multiplicative noise, only milstein and heun should work
    eqs = Equations("dv/dt = -v / (10*ms) + v*xi*second**-.5: 1")
    for integrator in (linear, independent, euler, exponential_euler, rk2, rk4):
        with pytest.raises(UnsupportedEquationsException):
            apply_stateupdater(eqs, variables, integrator)

    for integrator in (heun, milstein):
        with catch_logs() as logs:
            returned = apply_stateupdater(eqs, variables, method=integrator)
            assert (
                len(logs) == 0
            ), f"Got {len(logs)} unexpected warnings: {str([l[2] for l in logs])}"

    # Arbitrary functions (converting equations into abstract code) should
    # always work
    my_stateupdater = lambda eqs, vars, options: "x_new = x"
    with catch_logs() as logs:
        returned = apply_stateupdater(eqs, variables, method=my_stateupdater)
        # No warning here
        assert len(logs) == 0

    # Specification with names
    eqs = Equations("dv/dt = -v / (10*ms) : 1")
    for name, integrator in [
        ("exact", exact),
        ("linear", linear),
        ("euler", euler),
        # ('independent', independent), #TODO: Removed "independent" here due to the issue in sympy 0.7.4
        ("exponential_euler", exponential_euler),
        ("rk2", rk2),
        ("rk4", rk4),
        ("heun", heun),
        ("milstein", milstein),
    ]:
        with catch_logs() as logs:
            returned = apply_stateupdater(eqs, variables, method=name)
            # No warning here
            assert len(logs) == 0

    # Now all except heun and milstein should refuse to work
    eqs = Equations("dv/dt = -v / (10*ms) + v*xi*second**-.5: 1")
    for name in [
        "linear",
        "exact",
        "independent",
        "euler",
        "exponential_euler",
        "rk2",
        "rk4",
    ]:
        with pytest.raises(UnsupportedEquationsException):
            apply_stateupdater(eqs, variables, method=name)

    # milstein should work
    with catch_logs() as logs:
        apply_stateupdater(eqs, variables, method="milstein")
        assert len(logs) == 0

    # heun should work
    with catch_logs() as logs:
        apply_stateupdater(eqs, variables, method="heun")
        assert len(logs) == 0

    # non-existing name
    with pytest.raises(ValueError):
        apply_stateupdater(eqs, variables, method="does_not_exist")

    # Automatic state updater choice should return linear for linear equations,
    # euler for non-linear, non-stochastic equations and equations with
    # additive noise, heun for equations with multiplicative noise
    # Because it is somewhat fragile, the "independent" state updater is not
    # included in this list
    all_methods = ["linear", "exact", "exponential_euler", "euler", "heun", "milstein"]
    eqs = Equations("dv/dt = -v / (10*ms) : 1")
    with catch_logs(log_level=logging.INFO) as logs:
        apply_stateupdater(eqs, variables, all_methods)
        assert len(logs) == 1
        assert ("linear" in logs[0][2]) or ("exact" in logs[0][2])

    # This is conditionally linear
    eqs = Equations(
        """dv/dt = -(v + w**2)/ (10*ms) : 1
                       dw/dt = -w/ (10*ms) : 1"""
    )
    with catch_logs(log_level=logging.INFO) as logs:
        apply_stateupdater(eqs, variables, all_methods)
        assert len(logs) == 1
        assert "exponential_euler" in logs[0][2]

    # # Do not test for now
    # eqs = Equations('dv/dt = sin(t) / (10*ms) : 1')
    # assert apply_stateupdater(eqs, variables) is independent

    eqs = Equations("dv/dt = -sqrt(v) / (10*ms) : 1")
    with catch_logs(log_level=logging.INFO) as logs:
        apply_stateupdater(eqs, variables, all_methods)
        assert len(logs) == 1
        assert "'euler'" in logs[0][2]

    eqs = Equations("dv/dt = -v / (10*ms) + 0.1*second**-.5*xi: 1")
    with catch_logs(log_level=logging.INFO) as logs:
        apply_stateupdater(eqs, variables, all_methods)
        assert len(logs) == 1
        assert "'euler'" in logs[0][2]

    eqs = Equations("dv/dt = -v / (10*ms) + v*0.1*second**-.5*xi: 1")
    with catch_logs(log_level=logging.INFO) as logs:
        apply_stateupdater(eqs, variables, all_methods)
        assert len(logs) == 1
        assert "'heun'" in logs[0][2]


@pytest.mark.standalone_compatible
def test_subexpressions_basic():
    """
    Make sure that the integration of a (non-stochastic) differential equation
    does not depend on whether it's formulated using subexpressions.
    """
    # no subexpression
    eqs1 = "dv/dt = (-v + sin(2*pi*100*Hz*t)) / (10*ms) : 1"
    # same with subexpression
    eqs2 = """dv/dt = I / (10*ms) : 1
              I = -v + sin(2*pi*100*Hz*t): 1"""
    method = "euler"
    G1 = NeuronGroup(1, eqs1, method=method)
    G1.v = 1
    G2 = NeuronGroup(1, eqs2, method=method)
    G2.v = 1
    mon1 = StateMonitor(G1, "v", record=True)
    mon2 = StateMonitor(G2, "v", record=True)
    run(10 * ms)
    assert_equal(mon1.v, mon2.v, f"Results for method {method} differed!")


def test_subexpressions():
    """
    Make sure that the integration of a (non-stochastic) differential equation
    does not depend on whether it's formulated using subexpressions.
    """
    # no subexpression
    eqs1 = "dv/dt = (-v + sin(2*pi*100*Hz*t)) / (10*ms) : 1"
    # same with subexpression
    eqs2 = """dv/dt = I / (10*ms) : 1
              I = -v + sin(2*pi*100*Hz*t): 1"""

    methods = [
        "exponential_euler",
        "rk2",
        "rk4",
    ]  # euler is tested in test_subexpressions_basic
    for method in methods:
        G1 = NeuronGroup(1, eqs1, method=method)
        G1.v = 1
        G2 = NeuronGroup(1, eqs2, method=method)
        G2.v = 1
        mon1 = StateMonitor(G1, "v", record=True)
        mon2 = StateMonitor(G2, "v", record=True)
        net = Network(G1, mon1, G2, mon2)
        net.run(10 * ms)
        assert_equal(mon1.v, mon2.v, f"Results for method {method} differed!")


@pytest.mark.codegen_independent
def test_locally_constant_check():
    default_dt = defaultclock.dt
    # The linear state update can handle additive time-dependent functions
    # (e.g. a TimedArray) but only if it can be safely assumed that the function
    # is constant over a single time check
    ta0 = TimedArray(np.array([1]), dt=default_dt)  # ok
    ta1 = TimedArray(np.array([1]), dt=2 * default_dt)  # ok
    ta2 = TimedArray(np.array([1]), dt=default_dt / 2)  # not ok
    ta3 = TimedArray(np.array([1]), dt=default_dt * 1.5)  # not ok

    for ta_func, ok in zip([ta0, ta1, ta2, ta3], [True, True, False, False]):
        # additive
        G = NeuronGroup(
            1,
            "dv/dt = -v/(10*ms) + ta(t)*Hz : 1",
            method="exact",
            namespace={"ta": ta_func},
        )
        net = Network(G)
        if ok:
            # This should work
            net.run(0 * ms)
        else:
            # This should not
            with catch_logs():
                with pytest.raises(BrianObjectException) as exc:
                    net.run(0 * ms)
                    assert exc.errisinstance(UnsupportedEquationsException)

        # multiplicative
        G = NeuronGroup(
            1, "dv/dt = -v*ta(t)/(10*ms) : 1", method="exact", namespace={"ta": ta_func}
        )
        net = Network(G)
        if ok:
            # This should work
            net.run(0 * ms)
        else:
            # This should not
            with catch_logs():
                with pytest.raises(BrianObjectException) as exc:
                    net.run(0 * ms)
                    assert exc.errisinstance(UnsupportedEquationsException)

    # If the argument is more than just "t", we cannot guarantee that it is
    # actually locally constant
    G = NeuronGroup(
        1, "dv/dt = -v*ta(t/2.0)/(10*ms) : 1", method="exact", namespace={"ta": ta0}
    )
    net = Network(G)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms)
    assert exc_isinstance(exc, UnsupportedEquationsException)

    # Arbitrary functions are not constant over a time step
    G = NeuronGroup(1, "dv/dt = -v/(10*ms) + sin(2*pi*100*Hz*t)*Hz : 1", method="exact")
    net = Network(G)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms)
    assert exc_isinstance(exc, UnsupportedEquationsException)

    # Stateful functions aren't either
    G = NeuronGroup(1, "dv/dt = -v/(10*ms) + rand()*Hz : 1", method="exact")
    net = Network(G)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms)
    assert exc_isinstance(exc, UnsupportedEquationsException)

    # Neither is "t" itself
    G = NeuronGroup(1, "dv/dt = -v/(10*ms) + t/second**2 : 1", method="exact")
    net = Network(G)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms)
    assert exc_isinstance(exc, UnsupportedEquationsException)

    # But if the argument is not referring to t, all should be well
    G = NeuronGroup(
        1, "dv/dt = -v/(10*ms) + sin(2*pi*100*Hz*5*second)*Hz : 1", method="exact"
    )
    net = Network(G)
    net.run(0 * ms)


def test_refractory():
    # Compare integration with and without the addition of refractoriness --
    # note that the cell here is not spiking, so it should never be in the
    # refractory period and therefore the results should be exactly identical
    # with and without (unless refractory)
    eqs_base = "dv/dt = -v/(10*ms) : 1"
    for method in [
        "linear",
        "exact",
        "independent",
        "euler",
        "exponential_euler",
        "rk2",
        "rk4",
    ]:
        G_no_ref = NeuronGroup(10, eqs_base, method=method)
        G_no_ref.v = "(i+1)/11."
        G_ref = NeuronGroup(
            10, f"{eqs_base}(unless refractory)", refractory=1 * ms, method=method
        )
        G_ref.v = "(i+1)/11."
        net = Network(G_ref, G_no_ref)
        net.run(10 * ms)
        assert_allclose(
            G_no_ref.v[:],
            G_ref.v[:],
            err_msg="Results with and without refractoriness differ for method %s."
            % method,
        )


def test_refractory_stochastic(fake_randn_randn_fixture):
    eqs_base = "dv/dt = -v/(10*ms) + second**-.5*xi : 1"

    for method in ["euler", "heun", "milstein"]:
        G_no_ref = NeuronGroup(10, eqs_base, method=method)
        G_no_ref.v = "(i+1)/11."
        G_ref = NeuronGroup(
            10, f"{eqs_base} (unless refractory)", refractory=1 * ms, method=method
        )
        G_ref.v = "(i+1)/11."
        net = Network(G_ref, G_no_ref)
        net.run(10 * ms)
        assert_allclose(
            G_no_ref.v[:],
            G_ref.v[:],
            err_msg="Results with and without refractoriness differ for method %s."
            % method,
        )


@pytest.mark.standalone_compatible
def test_check_for_invalid_values_linear_integrator():
    # A differential equation that cannot be solved by the linear
    # integrator should return nan values to warn the user, and not silently
    # return incorrect values. See discussion on
    # https://github.com/brian-team/brian2/issues/626
    a = 0.0 / ms
    b = 1.0 / ms
    c = -0.5 / ms
    d = -0.1 / ms
    eqs = """
    dx/dt = a * x + b * y : 1
    dy/dt = c * x + d * y : 1
    """
    G = NeuronGroup(
        1,
        eqs,
        threshold="x > 100",
        reset="x = 0",
        method="exact",
        method_options={"simplify": False},
    )
    G.x = 1
    BrianLogger._log_messages.clear()  # because the log message is set to be shown only once
    with catch_logs() as clog:
        try:
            run(1 * ms)
            # this check allows for the possibility that we improve the linear
            # integrator in the future so that it can handle this equation
            if numpy.isnan(G.x[0]):
                assert "invalid_values" in repr(clog)
            else:
                assert G.x[0] != 0
        except BrianObjectException as exc:
            assert isinstance(exc.__cause__, UnsupportedEquationsException)


if __name__ == "__main__":
    import time

    from brian2 import prefs

    start = time.time()

    test_determination()
    test_explicit_stateupdater_parsing()
    test_non_autonomous_equations()
    test_str_repr()
    test_multiplicative_noise()
    test_multiple_noise_variables_basic()
    test_multiple_noise_variables_extended()
    test_temporary_variables()
    test_temporary_variables2()
    test_integrator_code()
    test_integrator_code2()
    test_illegal_calls()
    test_priority()
    test_registration()
    test_subexpressions()
    test_locally_constant_check()
    test_refractory()
    # # Need the fake random number generator from tests/conftest.py
    # test_refractory_stochastic()
    # test_multiple_noise_variables_deterministic_noise()
    # test_pure_noise_deterministic()
    test_check_for_invalid_values_linear_integrator()
    print("Tests took", time.time() - start)



"""
Package providing the code generation framework.
"""

# Import the runtime packages so that they can register their preferences

# isort:skip_file

from .runtime import *
from . import _prefs
from . import cpp_prefs as _cpp_prefs

__all__ = ["NumpyCodeObject", "CythonCodeObject"]



#!/usr/bin/env python
# coding: latin-1
"""
CUBA example with delays.
"""

import sys, time, os
from brian2 import *

standalone = int(sys.argv[-2])
n_threads  = int(sys.argv[-1])
path       = 'data_example_bis_%d' %n_threads

if standalone == 1:
    set_device('cpp_standalone')
    brian_prefs.codegen.cpp_standalone.openmp_threads = n_threads

start      = time.time()
n_cells    = 1000

numpy.random.seed(42)
connectivity = numpy.random.randn(n_cells, n_cells)

taum       = 20 * ms
taus       = 5 * ms
Vt         = -50 * mV
Vr         = -60 * mV
El         = -49 * mV

fac        = (60 * 0.27 / 10)  # excitatory synaptic weight (voltage)

gmax       = 20*fac
dApre      = .01
taupre     = 20 * ms
taupost    = taupre
dApost     = -dApre * taupre / taupost * 1.05
dApost    *=  0.1*gmax
dApre     *=  0.1*gmax


eqs  = Equations('''
dv/dt = (g-(v-El))/taum : volt
dg/dt = -g/taus         : volt
''')

P    = NeuronGroup(n_cells, model=eqs, threshold='v>Vt', reset='v=Vr', refractory=5 * ms)
P.v  = Vr + numpy.random.rand(len(P)) * (Vt - Vr)
P.g  = 0 * mV



S    = Synapses(P, P, 
                    model = '''dApre/dt=-Apre/taupre    : 1 (event-driven)    
                               dApost/dt=-Apost/taupost : 1 (event-driven)
                               w                        : 1''', 
                    pre = '''g     += w*mV
                             Apre  += dApre
                             w      = w + Apost''',
                    post = '''Apost += dApost
                              w      = w + Apre''')
S.connect()
S.w  = fac*connectivity.flatten()


spike_mon = SpikeMonitor(P)
state_mon = StateMonitor(S, 'w', record=range(10), when=Clock(dt=0.1*second))
v_mon     = StateMonitor(P, 'v', record=range(10))

if standalone == 1:
    device.insert_code('main', 'std::clock_t start = std::clock();')

run(5 * second, report='text')

if standalone == 1:
    device.insert_code('main', '''
        std::ofstream myfile ("speed.txt");
        if (myfile.is_open())
        {
            double value = (double) (std::clock() - start)/(%d * CLOCKS_PER_SEC); 
            myfile << value << std::endl;
            myfile.close();
        }
        ''' %(max(1, n_threads)))


try:
    os.removedirs(path)
except Exception:
    pass

if standalone == 1:
    device.build(project_dir=path, compile_project=True, run_project=True, debug=False)



#!/usr/bin/env python
"""
Input-Frequency curve of a IF model.

Network: 1000 unconnected integrate-and-fire neurons (leaky IF)
with an input parameter v0.
The input is set differently for each neuron.
"""
from brian2 import *

n = 1000
duration = 1*second
tau = 10*ms
eqs = '''
dv/dt = (v0 - v) / tau : volt (unless refractory)
v0 : volt
'''
group = NeuronGroup(n, eqs, threshold='v > 10*mV', reset='v = 0*mV',
                    refractory=5*ms, method='exact')
group.v = 0*mV
group.v0 = '20*mV * i / (n-1)'

monitor = SpikeMonitor(group)

run(duration)
plot(group.v0/mV, monitor.count / duration)
xlabel('v0 (mV)')
ylabel('Firing rate (sp/s)')
show()



{# USES_VARIABLES { N } #}
{% extends 'common_group.pyx' %}

{% block maincode %}
    {% set _target_var_array = get_array_name(_target_var) %}
    {% set _index_array = get_array_name(_index_var) %}
    cdef int _target_idx

    # Set all the target variable values to zero
    for _target_idx in range({{_target_size_name}}):
        {{_target_var_array}}[_target_idx + {{_target_start}}] = 0

    # scalar code
    _vectorisation_idx = 1
    {{scalar_code|autoindent}}

    for _idx in range({{N}}):
        # vector_code
        vectorisation_idx = _idx
        {{ vector_code | autoindent }}
        {{_target_var_array}}[{{_index_array}}[_idx]] += _synaptic_var

{% endblock %}



import os
import shutil
import sphinx
if not sphinx.version_info >= (1, 8):
    raise ImportError('Need sphinx version 1.8')
from sphinx.cmd.build import main as sphinx_main
import sys

os.chdir(os.path.join(os.path.dirname(__file__), '../../../docs_sphinx'))
if os.path.exists('../docs'):
    shutil.rmtree('../docs')
# Some code (e.g. the definition of preferences) might need to know that Brian
# is used to build the documentation. The READTHEDOCS variable is set
# on the readthedocs.io server automatically, so we reuse it here to signal
# a documentation build
os.environ['READTHEDOCS'] = 'True'
sys.exit(sphinx_main(['-b', 'html', '.', '../docs']))



r"""
A module providing some physical units as `Quantity` objects. Note that these
units are not imported by wildcard imports (e.g. `from brian2 import *`), they
have to be imported explicitly. You can use ``import ... as ...`` to import them
with shorter names, e.g.::

    from brian2.units.constants import faraday_constant as F

The available constants are:

==================== ================== ======================= ==================================================================
Constant             Symbol(s)          Brian name              Value
==================== ================== ======================= ==================================================================
Avogadro constant    :math:`N_A, L`     ``avogadro_constant``   :math:`6.022140857\times 10^{23}\,\mathrm{mol}^{-1}`
Boltzmann constant   :math:`k`          ``boltzmann_constant``  :math:`1.38064852\times 10^{-23}\,\mathrm{J}\,\mathrm{K}^{-1}`
Electric constant    :math:`\epsilon_0` ``electric_constant``   :math:`8.854187817\times 10^{-12}\,\mathrm{F}\,\mathrm{m}^{-1}`
Electron mass        :math:`m_e`        ``electron_mass``       :math:`9.10938356\times 10^{-31}\,\mathrm{kg}`
Elementary charge    :math:`e`          ``elementary_charge``   :math:`1.6021766208\times 10^{-19}\,\mathrm{C}`
Faraday constant     :math:`F`          ``faraday_constant``    :math:`96485.33289\,\mathrm{C}\,\mathrm{mol}^{-1}`
Gas constant         :math:`R`          ``gas_constant``        :math:`8.3144598\,\mathrm{J}\,\mathrm{mol}^{-1}\,\mathrm{K}^{-1}`
Magnetic constant    :math:`\mu_0`      ``magnetic_constant``   :math:`12.566370614\times 10^{-7}\,\mathrm{N}\,\mathrm{A}^{-2}`
Molar mass constant  :math:`M_u`        ``molar_mass_constant`` :math:`1\times 10^{-3}\,\mathrm{kg}\,\mathrm{mol}^{-1}`
0°C                                     ``zero_celsius``        :math:`273.15\,\mathrm{K}`
==================== ================== ======================= ==================================================================
"""

import numpy as np

from .allunits import (
    amp,
    coulomb,
    farad,
    gram,
    joule,
    kelvin,
    kilogram,
    meter,
    mole,
    newton,
)
from .fundamentalunits import Unit

Unit.automatically_register_units = False

#: Avogadro constant (http://physics.nist.gov/cgi-bin/cuu/Value?na)
avogadro_constant = 6.022140857e23 / mole
#: Boltzmann constant (physics.nist.gov/cgi-bin/cuu/Value?k)
boltzmann_constant = 1.38064852e-23 * joule / kelvin
#: electric constant (http://physics.nist.gov/cgi-bin/cuu/Value?ep0)
electric_constant = 8.854187817e-12 * farad / meter
#: Electron rest mass (physics.nist.gov/cgi-bin/cuu/Value?me)
electron_mass = 9.10938356e-31 * kilogram
#: Elementary charge (physics.nist.gov/cgi-bin/cuu/Value?e)
elementary_charge = 1.6021766208e-19 * coulomb
#: Faraday constant (http://physics.nist.gov/cgi-bin/cuu/Value?f)
faraday_constant = 96485.33289 * coulomb / mole
#: gas constant (http://physics.nist.gov/cgi-bin/cuu/Value?r)
gas_constant = 8.3144598 * joule / mole / kelvin
#: Magnetic constant (http://physics.nist.gov/cgi-bin/cuu/Value?mu0)
magnetic_constant = 4 * np.pi * 1e-7 * newton / amp**2
#: Molar mass constant (http://physics.nist.gov/cgi-bin/cuu/Value?mu)
molar_mass_constant = 1 * gram / mole
#: zero degree Celsius
zero_celsius = 273.15 * kelvin

Unit.automatically_register_units = True



"""
A short cylinder with constant injection at one end.
"""
from brian2 import *

defaultclock.dt = 0.01*ms

# Morphology
diameter = 1*um
length = 300*um
Cm = 1*uF/cm**2
Ri = 150*ohm*cm
N = 200
morpho = Cylinder(diameter=diameter, length=length, n=N)

# Passive channels
gL = 1e-4*siemens/cm**2
EL = -70*mV
eqs = '''
Im = gL * (EL - v) : amp/meter**2
I : amp (point current)
'''

neuron = SpatialNeuron(morphology=morpho, model=eqs, Cm=Cm, Ri=Ri,
                       method='exponential_euler')
neuron.v = EL

la = neuron.space_constant[0]
print("Electrotonic length: %s" % la)

neuron.I[0] = 0.02*nA # injecting at the left end
run(100*ms, report='text')

plot(neuron.distance/um, neuron.v/mV, 'kx')
# Theory
x = neuron.distance
ra = la * 4 * Ri / (pi * diameter**2)
theory = EL + ra * neuron.I[0] * cosh((length - x) / la) / sinh(length / la)
plot(x/um, theory/mV, 'r')
xlabel('x (um)')
ylabel('v (mV)')
show()



"""Extract reference documentation from the NumPy source tree.

"""

import inspect
import pydoc
import re
import textwrap
from warnings import warn

from sphinx.pycode import ModuleAnalyzer


class Reader:
    """A line-based string reader."""

    def __init__(self, data):
        """
        Parameters
        ----------
        data : str
           String with lines separated by '\n'.

        """
        if isinstance(data, list):
            self._str = data
        else:
            self._str = data.split("\n")  # store string as list of lines

        self.reset()

    def __getitem__(self, n):
        return self._str[n]

    def reset(self):
        self._l = 0  # current line nr

    def read(self):
        if not self.eof():
            out = self[self._l]
            self._l += 1
            return out
        else:
            return ""

    def seek_next_non_empty_line(self):
        for line in self[self._l :]:
            if line.strip():
                break
            else:
                self._l += 1

    def eof(self):
        return self._l >= len(self._str)

    def read_to_condition(self, condition_func):
        start = self._l
        for line in self[start:]:
            if condition_func(line):
                return self[start : self._l]
            self._l += 1
            if self.eof():
                return self[start : self._l + 1]
        return []

    def read_to_next_empty_line(self):
        self.seek_next_non_empty_line()

        def is_empty(line):
            return not line.strip()

        return self.read_to_condition(is_empty)

    def read_to_next_unindented_line(self):
        def is_unindented(line):
            return line.strip() and (len(line.lstrip()) == len(line))

        return self.read_to_condition(is_unindented)

    def peek(self, n=0):
        if self._l + n < len(self._str):
            return self[self._l + n]
        else:
            return ""

    def is_empty(self):
        return not "".join(self._str).strip()


class NumpyDocString:
    def __init__(self, docstring, config=None):
        docstring = textwrap.dedent(docstring).split("\n")

        self._doc = Reader(docstring)
        self._parsed_data = {
            "Signature": "",
            "Summary": [""],
            "Extended Summary": [],
            "Parameters": [],
            "Returns": [],
            "Raises": [],
            "Warns": [],
            "Other Parameters": [],
            "Attributes": [],
            "Methods": [],
            "See Also": [],
            "Notes": [],
            "Warnings": [],
            "References": "",
            "Examples": "",
            "index": {},
        }

        self._parse()

    def __getitem__(self, key):
        return self._parsed_data[key]

    def __setitem__(self, key, val):
        if key not in self._parsed_data:
            warn(f"Unknown section {key}")
        else:
            self._parsed_data[key] = val

    def _is_at_section(self):
        self._doc.seek_next_non_empty_line()

        if self._doc.eof():
            return False

        l1 = self._doc.peek().strip()  # e.g. Parameters

        if l1.startswith(".. index::"):
            return True

        l2 = self._doc.peek(1).strip()  # ---------- or ==========
        return l2.startswith("-" * len(l1)) or l2.startswith("=" * len(l1))

    def _strip(self, doc):
        start = stop = 0
        for i, line in enumerate(doc):
            if line.strip():
                start = i
                break

        for i, line in enumerate(doc[::-1]):
            if line.strip():
                stop = i
                break

        return doc[start : len(doc) - stop]

    def _read_to_next_section(self):
        section = self._doc.read_to_next_empty_line()

        while not self._is_at_section() and not self._doc.eof():
            if not self._doc.peek(-1).strip():  # previous line was empty
                section += [""]

            section += self._doc.read_to_next_empty_line()

        return section

    def _read_sections(self):
        while not self._doc.eof():
            data = self._read_to_next_section()
            name = data[0].strip()

            if name.startswith(".."):  # index section
                yield name, data[1:]
            elif len(data) < 2:
                yield StopIteration
            else:
                yield name, self._strip(data[2:])

    def _parse_param_list(self, content):
        r = Reader(content)
        params = []
        while not r.eof():
            header = r.read().strip()
            if " : " in header:
                arg_name, arg_type = header.split(" : ")[:2]
            else:
                arg_name, arg_type = header, ""

            desc = r.read_to_next_unindented_line()
            desc = dedent_lines(desc)

            params.append((arg_name, arg_type, desc))

        return params

    _name_rgx = re.compile(
        r"^\s*(:(?P<role>\w+):`(?P<name>[a-zA-Z0-9_.-]+)`|"
        r" (?P<name2>[a-zA-Z0-9_.-]+))\s*",
        re.X,
    )

    def _parse_see_also(self, content):
        """
        func_name : Descriptive text
            continued text
        another_func_name : Descriptive text
        func_name1, func_name2, :meth:`func_name`, func_name3

        """
        items = []

        def parse_item_name(text):
            """Match ':role:`name`' or 'name'"""
            m = self._name_rgx.match(text)
            if m:
                g = m.groups()
                if g[1] is None:
                    return g[3], None
                else:
                    return g[2], g[1]
            raise ValueError(f"{text} is not a item name")

        def push_item(name, rest):
            if not name:
                return
            name, role = parse_item_name(name)
            items.append((name, list(rest), role))
            del rest[:]

        current_func = None
        rest = []

        for line in content:
            if not line.strip():
                continue

            m = self._name_rgx.match(line)
            if m and line[m.end() :].strip().startswith(":"):
                push_item(current_func, rest)
                current_func, line = line[: m.end()], line[m.end() :]
                rest = [line.split(":", 1)[1].strip()]
                if not rest[0]:
                    rest = []
            elif not line.startswith(" "):
                push_item(current_func, rest)
                current_func = None
                if "," in line:
                    for func in line.split(","):
                        if func.strip():
                            push_item(func, [])
                elif line.strip():
                    current_func = line
            elif current_func is not None:
                rest.append(line.strip())
        push_item(current_func, rest)
        return items

    def _parse_index(self, section, content):
        """
        .. index: default
           :refguide: something, else, and more

        """

        def strip_each_in(lst):
            return [s.strip() for s in lst]

        out = {}
        section = section.split("::")
        if len(section) > 1:
            out["default"] = strip_each_in(section[1].split(","))[0]
        for line in content:
            line = line.split(":")
            if len(line) > 2:
                out[line[1]] = strip_each_in(line[2].split(","))
        return out

    def _parse_summary(self):
        """Grab signature (if given) and summary"""
        if self._is_at_section():
            return

        summary = self._doc.read_to_next_empty_line()
        summary_str = " ".join([s.strip() for s in summary]).strip()
        if re.compile(r"^([\w., ]+=)?\s*[\w.]+\(.*\)$").match(summary_str):
            self["Signature"] = summary_str
            if not self._is_at_section():
                self["Summary"] = self._doc.read_to_next_empty_line()
        else:
            self["Summary"] = summary

        if not self._is_at_section():
            self["Extended Summary"] = self._read_to_next_section()

    def _parse(self):
        self._doc.reset()
        self._parse_summary()

        for section, content in self._read_sections():
            if not section.startswith(".."):
                section = " ".join([s.capitalize() for s in section.split(" ")])
            if section in (
                "Parameters",
                "Returns",
                "Raises",
                "Warns",
                "Other Parameters",
                "Attributes",
                "Methods",
            ):
                self[section] = self._parse_param_list(content)
            elif section.startswith(".. index::"):
                self["index"] = self._parse_index(section, content)
            elif section == "See Also":
                self["See Also"] = self._parse_see_also(content)
            else:
                self[section] = content

    # string conversion routines

    def _str_header(self, name, symbol="-"):
        return [name, len(name) * symbol]

    def _str_indent(self, doc, indent=4):
        out = []
        for line in doc:
            out += [" " * indent + line]
        return out

    def _str_signature(self):
        if self["Signature"]:
            return [self["Signature"].replace("*", r"\*")] + [""]
        else:
            return [""]

    def _str_summary(self):
        if self["Summary"]:
            return self["Summary"] + [""]
        else:
            return []

    def _str_extended_summary(self):
        if self["Extended Summary"]:
            return self["Extended Summary"] + [""]
        else:
            return []

    def _str_param_list(self, name):
        out = []
        if self[name]:
            out += self._str_header(name)
            for param, param_type, desc in self[name]:
                out += [f"{param} : {param_type}"]
                out += self._str_indent(desc)
            out += [""]
        return out

    def _str_section(self, name):
        out = []
        if self[name]:
            out += self._str_header(name)
            out += self[name]
            out += [""]
        return out

    def _str_see_also(self, func_role):
        if not self["See Also"]:
            return []
        out = []
        out += self._str_header("See Also")
        last_had_desc = True
        for func, desc, role in self["See Also"]:
            if role:
                link = f":{role}:`{func}`"
            elif func_role:
                link = f":{func_role}:`{func}`"
            else:
                link = f"`{func}`_"
            if desc or last_had_desc:
                out += [""]
                out += [link]
            else:
                out[-1] += f", {link}"
            if desc:
                out += self._str_indent([" ".join(desc)])
                last_had_desc = True
            else:
                last_had_desc = False
        out += [""]
        return out

    def _str_index(self):
        idx = self["index"]
        out = []
        out += [f".. index:: {idx.get('default', '')}"]
        for section, references in idx.items():
            if section == "default":
                continue
            out += [f"   :{section}: {', '.join(references)}"]
        return out

    def __str__(self, func_role=""):
        out = []
        out += self._str_signature()
        out += self._str_summary()
        out += self._str_extended_summary()
        for param_list in (
            "Parameters",
            "Returns",
            "Other Parameters",
            "Raises",
            "Warns",
        ):
            out += self._str_param_list(param_list)
        out += self._str_section("Warnings")
        out += self._str_see_also(func_role)
        for s in ("Notes", "References", "Examples"):
            out += self._str_section(s)
        for param_list in ("Attributes", "Methods"):
            out += self._str_param_list(param_list)
        out += self._str_index()
        return "\n".join(out)


def indent(str, indent=4):
    indent_str = " " * indent
    if str is None:
        return indent_str
    lines = str.split("\n")
    return "\n".join(indent_str + line for line in lines)


def dedent_lines(lines):
    """Deindent a list of lines maximally"""
    return textwrap.dedent("\n".join(lines)).split("\n")


def header(text, style="-"):
    return f"{text}\n{style * len(text)}\n"


class FunctionDoc(NumpyDocString):
    def __init__(self, func, role="func", doc=None, config=None):
        self._f = func
        self._role = role  # e.g. "func" or "meth"

        if doc is None:
            if func is None:
                raise ValueError("No function or docstring given")
            doc = inspect.getdoc(func) or ""
        NumpyDocString.__init__(self, doc)

        if not self["Signature"] and func is not None:
            func, func_name = self.get_func()
            try:
                # try to read signature
                argspec = str(inspect.signature(func))
                argspec = argspec.replace("*", r"\*")
                signature = f"{func_name}{argspec}"
            except (TypeError, ValueError):
                signature = f"{func_name}()"
            self["Signature"] = signature

    def get_func(self):
        func_name = getattr(self._f, "__name__", self.__class__.__name__)
        if inspect.isclass(self._f):
            if callable(self._f):
                func = self._f.__call__
            else:
                func = self._f.__init__
        else:
            func = self._f
        return func, func_name

    def __str__(self):
        out = ""

        _, func_name = self.get_func()

        roles = {"func": "function", "meth": "method"}

        if self._role:
            if self._role not in roles:
                print(f"Warning: invalid role {self._role}")
            out += f".. {roles.get(self._role, '')}:: {func_name}\n    \n\n"

        out += super().__str__(func_role=self._role)
        return out


class ClassDoc(NumpyDocString):
    extra_public_methods = ["__call__"]

    def __init__(self, cls, doc=None, modulename="", func_doc=FunctionDoc, config=None):
        if not inspect.isclass(cls) and cls is not None:
            raise ValueError(f"Expected a class or None, but got {cls!r}")
        self._cls = cls

        if modulename and not modulename.endswith("."):
            modulename += "."
        self._mod = modulename

        if doc is None:
            if cls is None:
                raise ValueError("No class or documentation string given")
            doc = pydoc.getdoc(cls)

        NumpyDocString.__init__(self, doc)

        if not self["Methods"]:
            self["Methods"] = [(name, "", "") for name in sorted(self.methods)]
        if not self["Attributes"]:
            self["Attributes"] = [(name, "", "") for name in sorted(self.properties)]

    @property
    def methods(self):
        if self._cls is None:
            return []
        methods = [
            name
            for name, func in self._cls.__dict__.items()
            if (
                (not name.startswith("_") or name in self.extra_public_methods)
                and (
                    (callable(func) and not isinstance(func, type))
                    or inspect.ismethoddescriptor(func)
                )
            )
        ]
        return methods

    @property
    def properties(self):
        if self._cls is None:
            return []
        analyzer = ModuleAnalyzer.for_module(self._cls.__module__)
        instance_members = {
            attr_name
            for (class_name, attr_name) in analyzer.find_attr_docs().keys()
            if class_name == self._cls.__name__
        }
        class_members = {
            name
            for name, func in self._cls.__dict__.items()
            if not name.startswith("_")
            and (func is None or inspect.isdatadescriptor(func))
        }

        return instance_members | class_members



"""
Input-Frequency curve of a HH model.

Network: 100 unconnected Hodgin-Huxley neurons with an input current I.
The input is set differently for each neuron.

This simulation should use exponential Euler integration.
"""
from brian2 import *

num_neurons = 100
duration = 2*second

# Parameters
area = 20000*umetre**2
Cm = 1*ufarad*cm**-2 * area
gl = 5e-5*siemens*cm**-2 * area
El = -65*mV
EK = -90*mV
ENa = 50*mV
g_na = 100*msiemens*cm**-2 * area
g_kd = 30*msiemens*cm**-2 * area
VT = -63*mV

# The model
eqs = Equations('''
dv/dt = (gl*(El-v) - g_na*(m*m*m)*h*(v-ENa) - g_kd*(n*n*n*n)*(v-EK) + I)/Cm : volt
dm/dt = 0.32*(mV**-1)*4*mV/exprel((13.*mV-v+VT)/(4*mV))/ms*(1-m)-0.28*(mV**-1)*5*mV/exprel((v-VT-40.*mV)/(5*mV))/ms*m : 1
dn/dt = 0.032*(mV**-1)*5*mV/exprel((15.*mV-v+VT)/(5*mV))/ms*(1.-n)-.5*exp((10.*mV-v+VT)/(40.*mV))/ms*n : 1
dh/dt = 0.128*exp((17.*mV-v+VT)/(18.*mV))/ms*(1.-h)-4./(1+exp((40.*mV-v+VT)/(5.*mV)))/ms*h : 1
I : amp
''')
# Threshold and refractoriness are only used for spike counting
group = NeuronGroup(num_neurons, eqs,
                    threshold='v > -40*mV',
                    refractory='v > -40*mV',
                    method='exponential_euler')
group.v = El
group.I = '0.7*nA * i / num_neurons'

monitor = SpikeMonitor(group)

run(duration)

plot(group.I/nA, monitor.count / duration)
xlabel('I (nA)')
ylabel('Firing rate (sp/s)')
show()



"""
A dummy package to allow importing numpy and the unit-aware replacements of
numpy functions without having to know which functions are overwritten.

This can be used for example as ``import brian2.numpy_ as np``
"""

# isort:skip_file
# flake8: noqa

from numpy import *
from brian2.units.unitsafefunctions import *

# These will not be imported with a wildcard import to not overwrite the
# builtin names (mimicking the numpy behaviour)
from builtins import bool, float, complex

from numpy import round, abs, max, min

import numpy
import brian2.units.unitsafefunctions as brian2_functions

__all__ = []
__all__.extend(numpy.__all__)
__all__.extend(brian2_functions.__all__)



"""
Brian global preferences are stored as attributes of a `BrianGlobalPreferences`
object ``prefs``.
"""

import os
import re
from collections.abc import MutableMapping
from io import StringIO

from brian2.units.fundamentalunits import Quantity, have_same_dimensions
from brian2.utils.stringtools import deindent, indent

__all__ = ["PreferenceError", "BrianPreference", "prefs", "brian_prefs"]


def parse_preference_name(name):
    """
    Split a preference name into a base and end name.

    Parameters
    ----------
    name : str
        The full name of the preference.

    Returns
    -------
    basename : str
        The first part of the name up to the final ``.``.
    endname : str
        The last part of the name from the final ``.`` onwards.

    Examples
    --------
    >>> parse_preference_name('core.default_float_dtype')
    ('core', 'default_float_dtype')
    >>> parse_preference_name('codegen.cpp.compiler')
    ('codegen.cpp', 'compiler')
    """
    # parse the name
    parts = name.split(".")
    basename = ".".join(parts[:-1])
    endname = parts[-1]
    return basename, endname


def check_preference_name(name):
    """
    Make sure that a preference name is valid. This currently checks that the
    name does not contain illegal characters and does not clash with method
    names such as "keys" or "items".

    Parameters
    ----------
    name : str
        The name to check.

    Raises
    ------
    PreferenceError
        In case the name is invalid.
    """
    if not re.match("[A-Za-z][_a-zA-Z0-9]*$", name):
        raise PreferenceError(
            f"Illegal preference name '{name}': A preference "
            "name can only start with a letter and only "
            "contain letters, digits or underscore."
        )
    if name in dir(MutableMapping) or name in prefs.__dict__:
        raise PreferenceError(
            f"Illegal preference name '{name}': This is also the name of a method."
        )


class PreferenceError(Exception):
    """
    Exception relating to the Brian preferences system.
    """

    pass


class DefaultValidator:
    """
    Default preference validator

    Used by `BrianPreference` as the default validator if none is given.
    First checks if the provided value is of the same class as the default
    value, and then if the default is a `Quantity`, checks that the units
    match.
    """

    def __init__(self, value):
        self.value = value

    def __call__(self, value):
        if not isinstance(value, self.value.__class__):
            return False
        if isinstance(self.value, Quantity):
            if not have_same_dimensions(self.value, value):
                return False
        return True


class BrianPreference:
    """
    Used for defining a Brian preference.

    Parameters
    ----------
    default : object
        The default value.
    docs : str
        Documentation for the preference value.
    validator : func
        A function that True or False depending on whether the preference value
        is valid or not. If not specified, uses the `DefaultValidator` for the
        default value provided (check if the class is the same, and for
        `Quantity` objects, whether the units are consistent).
    representor : func
        A function that returns a string representation of a valid preference
        value that can be passed to `eval`. By default, uses `repr` which
        works in almost all cases.
    """

    def __init__(self, default, docs, validator=None, representor=repr):
        self.representor = representor
        if validator is None:
            validator = DefaultValidator(default)
        self.validator = validator
        self.default = default
        self.docs = docs


class BrianGlobalPreferences(MutableMapping):
    """
    Class of the ``prefs`` object.

    Used for getting/setting/validating/registering preference values.
    All preferences must be registered via `register_preferences`. To get or
    set a preference, you can either use a dictionary-based or an
    attribute-based interface::

        prefs['core.default_float_dtype'] = float32
        prefs.core.default_float_dtype = float32

    Preferences can be read from files, see `load_preferences` and
    `read_preference_file`. Note that `load_preferences` is called
    automatically when Brian has finished importing.
    """

    def __init__(self):
        self.prefs = {}
        self.backup_prefs = {}
        self.prefs_unvalidated = {}
        self.pref_register = {}
        self.eval_namespace = {}
        exec(
            deindent(
                """
            from numpy import *
            from brian2.units import *
            from brian2.units.stdunits import *
            """
            ),
            self.eval_namespace,
        )

    def __getitem__(self, item):
        if item in self.pref_register:
            # This asks for a category, not a single preference
            return BrianGlobalPreferencesView(item, self)
        return self.prefs[item]

    def __len__(self):
        return len(self.prefs)

    def __iter__(self):
        return iter(self.prefs)

    def __contains__(self, item):
        return item in self.prefs

    def __setitem__(self, name, value):
        basename, endname = parse_preference_name(name)
        if basename not in self.pref_register:
            raise PreferenceError(
                "Preference category " + basename + " is unregistered. Spelling error?"
            )
        prefdefs, _ = self.pref_register[basename]
        if endname in prefdefs:
            # do validation
            pref = prefdefs[endname]
            if not pref.validator(value):
                raise PreferenceError(
                    f"Value '{value}' for preference '{name}' is invalid."
                )
            self.prefs[name] = value
            if name in self.prefs_unvalidated:
                del self.prefs_unvalidated[name]
        else:
            raise PreferenceError(
                f"Preference '{name}' is unregistered. Spelling error?"
            )

    def __delitem__(self, item):
        raise PreferenceError("Preferences cannot be deleted.")

    def __getattr__(self, name):
        if name in self.__dict__ or name.startswith("__"):
            return super().__getattribute__(name)

        # This function might get called from BrianGlobalPreferencesView with
        # a prefixed name -- therefore the name can contain dots!
        if name in self.pref_register:
            # This asks for a category, not a single preference
            return BrianGlobalPreferencesView(name, self)

        basename, _ = parse_preference_name(name)
        if len(basename) and basename not in self.pref_register:
            raise AssertionError(
                f"__getattr__ received basename '{basename}' which is "
                "unregistered. This should never happen!"
            )

        try:
            return self[name]
        except KeyError as ex:
            raise AttributeError(
                f"Object of type {type(self).__name__} does not have an attribute '{ex.args[0]}'",
                name=ex.args[0],
                obj=self,
            ) from ex

    def __setattr__(self, name, value):
        # Do not allow to set a category name to something else
        if "pref_register" in self.__dict__ and name in self.pref_register:
            raise PreferenceError("Cannot set a preference category.")
        else:
            MutableMapping.__setattr__(self, name, value)

    def __delattr__(self, name):
        if "pref_register" in self.__dict__ and name in self.pref_register:
            raise PreferenceError("Cannot delete a preference category.")
        else:
            MutableMapping.__delattr__(self, name)

    toplevel_categories = property(
        fget=lambda self: [
            category for category in self.pref_register if "." not in category
        ],
        doc="The toplevel preference categories",
    )

    def _get_docstring(self):
        """
        Document the toplevel categories, used as a docstring for the object.
        """
        s = "Preference categories:\n\n"
        for category in self.toplevel_categories:
            s += "** %s **\n" % category
            _, category_doc = self.pref_register[category]
            s += "    " + category_doc + "\n\n"

        return s

    def __dir__(self):
        res = dir(type(self)) + list(self.__dict__)
        categories = self.toplevel_categories
        res.extend(categories)
        return res

    def eval_pref(self, value):
        """
        Evaluate a string preference in the units namespace
        """
        return eval(value, self.eval_namespace)

    def _set_preference(self, name, value):
        """
        Try to set the preference and allow for unregistered base names. This
        method is used internally when reading preferences from the file
        because the preferences are potentially defined in packages that are
        not imported yet. Unvalidated preferences are safed and will be
        validated as soon as the category is registered. `Network.run` will
        also check for unvalidated preferences.
        """
        basename, _ = parse_preference_name(name)
        if basename not in self.pref_register:
            self.prefs_unvalidated[name] = value
        else:
            # go via the standard __setitem__ method
            self[name] = value

    def _backup(self):
        """
        Store a backup copy of the preferences to restore with `_restore`.
        """
        self.backup_prefs.update(**self.prefs)

    def _restore(self):
        """
        Restore a copy of the values of the preferences backed up with `_backup`.
        """
        self.prefs.update(**self.backup_prefs)

    def _get_one_documentation(self, basename, link_targets):
        """
        Document a single category of preferences.
        """

        s = ""
        if basename not in self.pref_register:
            raise ValueError(
                f"No preferences under the name '{basename}' are registered"
            )
        prefdefs, basedoc = self.pref_register[basename]
        s += deindent(basedoc, docstring=True).strip() + "\n\n"
        for name in sorted(prefdefs.keys()):
            pref = prefdefs[name]
            name = basename + "." + name
            linkname = name.replace("_", "-").replace(".", "-")
            if link_targets:
                # Make a link target
                s += f".. _brian-pref-{linkname}:\n\n"
            s += f"``{name}`` = ``{pref.representor(pref.default)}``\n"
            s += indent(deindent(pref.docs, docstring=True))
            s += "\n\n"
        return s

    def get_documentation(self, basename=None, link_targets=True):
        """
        Generates a string documenting all preferences with the given
        `basename`. If no `basename` is given, all preferences are documented.
        """
        s = ""
        if basename is None:
            basenames = sorted(
                [tuple(basename.split(".")) for basename in self.pref_register]
            )
            for basename in basenames:
                lev = len(basename)
                basename = ".".join(basename)
                if lev == 1:
                    s += basename + "\n" + '"' * len(basename) + "\n\n"
                else:
                    s += "**" + basename + "**\n\n"
                s += self._get_one_documentation(basename, link_targets)
            # for basename in self.pref_register:
            # s += '**' + basename + '**\n\n'
            # s += basename+'\n'+'"'*len(basename)+'\n\n'
            # s += self._get_one_documentation(basename, link_targets)
        else:
            s += self._get_one_documentation(basename, link_targets)

        return s

    def _as_pref_file(self, valuefunc):
        """
        Helper function used to generate the preference file for the default or current preference values.
        """
        s = ""
        for basename, (prefdefs, basedoc) in self.pref_register.items():
            s += "#" + "-" * 79 + "\n"
            s += (
                "\n".join(
                    [
                        "# " + line
                        for line in deindent(basedoc, docstring=True)
                        .strip()
                        .split("\n")
                    ]
                )
                + "\n"
            )
            s += "#" + "-" * 79 + "\n\n"
            s += "[" + basename + "]\n\n"
            for name in sorted(prefdefs.keys()):
                pref = prefdefs[name]
                s += (
                    "\n".join(
                        [
                            "# " + line
                            for line in deindent(pref.docs, docstring=True)
                            .strip()
                            .split("\n")
                        ]
                    )
                    + "\n\n"
                )
                s += (
                    name
                    + " = "
                    + pref.representor(valuefunc(pref, basename + "." + name))
                    + "\n\n"
                )
        return s

    def _get_defaults_as_file(self):
        return self._as_pref_file(lambda pref, fullname: pref.default)

    defaults_as_file = property(
        fget=_get_defaults_as_file,
        doc="Get a Brian preference doc file format string for the default preferences",
    )

    def _get_as_file(self):
        return self._as_pref_file(lambda pref, fullname: self[fullname])

    as_file = property(
        fget=_get_as_file,
        doc="Get a Brian preference doc file format string for the current preferences",
    )

    def read_preference_file(self, file):
        """
        Reads a Brian preferences file

        The file format for Brian preferences is a plain text file of the form::

            a.b.c = 1
            # Comment line
            [a]
            b.d = 2
            [a.b]
            e = 3

        Blank and comment lines are ignored, all others should be of one of the
        following two forms::

            key = value
            [section]

        `eval` is called on the values, so strings should be written as, e.g.
        ``'3'`` rather than ``3``. The eval is called with all unit names
        available.
        Within a section, the section name is prepended to the key. So in the above
        example, it would give the following unvalidated dictionary::

            {'a.b.c': 1,
             'a.b.d': 2,
             'a.b.e': 3,
             }

        Parameters
        ----------
        file : file, str
            The file object or filename of the preference file.
        """
        if isinstance(file, str):
            filename = file
            file = open(file)
        else:
            filename = repr(file)
        lines = file.readlines()
        file.close()
        # remove empty lines
        lines = [line.strip() for line in lines]
        lines = [line for line in lines if line]
        # Remove comments
        lines = [line for line in lines if not line.startswith("#")]
        bases = []  # start with no base
        for line in lines:
            # Match section names, which are used as a prefix for subsequent entries
            m = re.match(r"\[([^\]]*)\]", line)
            if m:
                bases = m.group(1).strip().split(".")
                continue
            # Match entries
            m = re.match("(.*?)=(.*)", line)
            if m:
                extname = m.group(1).strip()
                value = m.group(2).strip()
                keyname = ".".join(bases + extname.split("."))
                self._set_preference(keyname, self.eval_pref(value))
                continue
            # Otherwise raise a parsing error
            raise PreferenceError("Parsing error in preference file " + filename)

    def load_preferences(self):
        """
        Load all the preference files, but do not validate them.

        Preference files are read in the following order:

        1. ``~/.brian/user_preferences`` from the user's home directory
        2. ``./brian_preferences`` from the current directory

        Files that are missing are ignored. Preferences read at each step
        override preferences from previous steps.

        See Also
        --------
        read_preference_file
        """
        user_dir = os.path.join(os.path.expanduser("~"), ".brian")
        user_prefs = os.path.join(user_dir, "user_preferences")
        cur_prefs = "brian_preferences"

        files = [user_prefs, cur_prefs]
        for file in files:
            try:
                self.read_preference_file(file)
            except OSError:
                pass

        # The "default_preferences" file is no longer used, but we raise a
        # warning if it is present (note that we do this after reading the
        # preference files, since they can affect the preferences of the logger
        # itself)
        curdir, _ = os.path.split(__file__)
        basedir = os.path.normpath(os.path.join(curdir, ".."))
        default_prefs = os.path.join(basedir, "default_preferences")
        if os.path.exists(default_prefs):
            from brian2.utils.logger import get_logger

            logger = get_logger(__name__)
            logger.warn(
                "Brian no longer loads preferences from the "
                f"'default_preferences' file (in '{basedir}'). Use a "
                "'user_preferences' file in "
                f"'{user_dir}', "
                "or a 'brian_preferences' file in the current "
                "directory instead.",
                name_suffix="deprecated_default_preferences",
                once=True,
            )

    def reset_to_defaults(self):
        """
        Resets the parameters to their default values.
        """
        self.read_preference_file(StringIO(self.defaults_as_file))

    def register_preferences(self, prefbasename, prefbasedoc, **prefs):
        """
        Registers a set of preference names, docs and validation functions.

        Parameters
        ----------
        prefbasename : str
            The base name of the preference.
        prefbasedoc : str
            Documentation for this base name
        **prefs : dict of (name, `BrianPreference`) pairs
            The preference names to be defined. The full preference name will be
            ``prefbasename.name``, and the `BrianPreference` value is used to define
            the default value, docs, and validation function.

        Raises
        ------
        PreferenceError
            If the base name is already registered.

        See Also
        --------
        BrianPreference
        """
        if prefbasename in self.pref_register:
            # During the initial import phase the same base category may be
            # created twice, ignore that
            previous = self.pref_register[prefbasename]
            if not (len(previous[0]) == 0 and previous[1] == prefbasedoc):
                raise PreferenceError(
                    "Base name " + prefbasename + " already registered."
                )
        # Check that the new category does not clash with a preference name of
        # the parent category. For example, if a category "a" with the
        # preference "b" is already registered, do not allow to register a
        # preference category "a.b"
        basename, category_name = parse_preference_name(prefbasename)
        if len(basename) and basename in self.pref_register:
            parent_preferences, _ = self.pref_register[basename]
            if category_name in parent_preferences:
                raise PreferenceError(
                    f"Cannot register category '{prefbasename}', "
                    f"parent category '{basename}' already has a "
                    f"preference named '{category_name}'."
                )

        self.pref_register[prefbasename] = (prefs, prefbasedoc)
        for k, v in prefs.items():
            fullname = prefbasename + "." + k
            # The converse of the above check: Check that a preference name
            # does not clash with a category
            if fullname in self.pref_register:
                raise PreferenceError(
                    f"Cannot register '{fullname}' as a preference, "
                    "it is already registered as a "
                    "preference category."
                )
            check_preference_name(k)
            self.prefs_unvalidated[fullname] = v.default
        self.do_validation()

        # Update the docstring (a new toplevel category might have been added)
        self.__doc__ = self._get_docstring()

    def do_validation(self):
        """
        Validates preferences that have not yet been validated.
        """
        for name, value in dict(self.prefs_unvalidated).items():
            self[name] = value

    def check_all_validated(self):
        """
        Checks that all preferences that have been set have been validated.

        Logs a warning if not. Should be called by `Network.run` or other
        key Brian functions.
        """
        if len(self.prefs_unvalidated):
            from brian2.utils.logger import get_logger

            logger = get_logger(__name__)
            logger.warn(
                "The following preferences values have been set but "
                "are not registered preferences:\n%s\nThis is usually "
                "because of a spelling mistake or missing library "
                "import." % ", ".join(self.prefs_unvalidated),
                once=True,
            )

    def __repr__(self):
        description = "<{classname} with top-level categories: {categories}>"
        categories = ", ".join(
            ['"%s"' % category for category in self.toplevel_categories]
        )
        return description.format(
            classname=self.__class__.__name__, categories=categories
        )


class BrianGlobalPreferencesView(MutableMapping):
    """
    A class allowing for accessing preferences in a subcategory. It forwards
    requests to `BrianGlobalPreferences` and provides documentation and
    autocompletion support for all preferences in the given category. This
    object is used to allow accessing preferences via attributes of the
    `prefs` object.

    Parameters
    ----------
    basename : str
        The name of the preference category. Has to correspond to a key in
        `BrianGlobalPreferences.pref_register`.
    all_prefs : `BrianGlobalPreferences`
        A reference to the main object storing the preferences.
    """

    def __init__(self, basename, all_prefs):
        self._basename = basename
        self._all_prefs = all_prefs
        self._subcategories = [
            key for key in all_prefs.pref_register if key.startswith(basename + ".")
        ]
        self._preferences = list(all_prefs.pref_register[basename][0].keys())
        self.__doc__ = all_prefs.get_documentation(
            basename=basename, link_targets=False
        )

    _sub_preferences = property(
        lambda self: [
            pref[len(self._basename + ".") :]
            for pref in self._all_prefs
            if pref.startswith(self._basename + ".")
        ],
        doc="All preferences in this category and its subcategories",
    )

    def __getitem__(self, item):
        return self._all_prefs[self._basename + "." + item]

    def __setitem__(self, item, value):
        self._all_prefs[self._basename + "." + item] = value

    def __delitem__(self, item):
        raise PreferenceError("Preferences cannot be deleted.")

    def __len__(self):
        return len(self._sub_preferences)

    def __iter__(self):
        return iter(self._sub_preferences)

    def __contains__(self, item):
        return item in self._sub_preferences

    def __getattr__(self, name):
        return getattr(self._all_prefs, self._basename + "." + name)

    def __setattr__(self, name, value):
        # Names starting with an underscore are not preferences but normal
        # instance attributes
        if name.startswith("_"):
            MutableMapping.__setattr__(self, name, value)
        else:
            self._all_prefs[self._basename + "." + name] = value

    def __delattr__(self, name):
        # Names starting with an underscore are not preferences but normal
        # instance attributes
        if name.startswith("_"):
            MutableMapping.__delattr__(self, name)
        else:
            del self._all_prefs[self._basename + "." + name]

    def __dir__(self):
        res = dir(type(self)) + list(self.__dict__)
        res.extend(self._preferences)
        res.extend(
            [category[len(self._basename + ".") :] for category in self._subcategories]
        )
        return res

    def __repr__(self):
        description = '<{classname} for preference category "{category}">'
        return description.format(
            classname=self.__class__.__name__, category=self._basename
        )


# : Object storing Brian's preferences
prefs = BrianGlobalPreferences()


# Simple class to give a useful error message when using `brian_prefs`
class ErrorRaiser:
    def __getattr__(self, item):
        raise AttributeError(
            "The global preferences object has been renamed "
            "from 'brian_prefs' to 'prefs'"
        )

    def __getitem__(self, item):
        raise AttributeError(
            "The global preferences object has been renamed "
            "from 'brian_prefs' to 'prefs'"
        )


brian_prefs = ErrorRaiser()



"""
This example shows how to run several, independent simulations in standalone mode.
Given that this example only involves a single neuron, an alternative – and arguably more elegant – solution
would be to run the simulations in a single `NeuronGroup`, where each neuron receives input with a different rate.

The example is a standalone equivalent of the one presented in :doc:`/tutorials/3-intro-to-brian-simulations`.
"""
import numpy as np
import matplotlib.pyplot as plt
import brian2 as b2
from time import time

b2.set_device('cpp_standalone', build_on_run=False)

if __name__ == "__main__":
    start_time = time()
    num_inputs = 100
    input_rate = 10 * b2.Hz
    weight = 0.1

    net = b2.Network()
    P = b2.PoissonGroup(num_inputs, rates=input_rate)
    eqs = """
    dv/dt = -v/tau : 1
    tau : second (constant)
    """
    G = b2.NeuronGroup(1, eqs, threshold='v>1', reset='v=0', method='euler')
    S = b2.Synapses(P, G, on_pre='v += weight')
    S.connect()
    M = b2.SpikeMonitor(G)
    net.add([P, G, S, M])

    net.run(1000 * b2.ms)

    b2.device.build(run=False)  # compile the code, but don't run it yet

    npoints = 15
    tau_range = np.linspace(1, 15, npoints) * b2.ms    

    output_rates = np.zeros(npoints)
    for ii in range(npoints):
        tau_i = tau_range[ii]
        b2.device.run(run_args={G.tau: tau_i})
        output_rates[ii] = M.num_spikes / b2.second

    print(f"Done in {time() - start_time}")

    plt.plot(tau_range/b2.ms, output_rates)
    plt.xlabel(r"$\tau$ (ms)")
    plt.ylabel("Firing rate (sp/s)")
    plt.show()



"""
Brian 2
"""

import logging
import signal


def _check_dependencies():
    """Check basic dependencies"""
    import sys

    missing = []
    try:
        import numpy
    except ImportError as ex:
        sys.stderr.write(f"Importing numpy failed: '{ex}'\n")
        missing.append("numpy")
    try:
        import sympy
    except ImportError as ex:
        sys.stderr.write(f"Importing sympy failed: '{ex}'\n")
        missing.append("sympy")
    try:
        import pyparsing
    except ImportError as ex:
        sys.stderr.write(f"Importing pyparsing failed: '{ex}'\n")
        missing.append("pyparsing")
    try:
        import jinja2
    except ImportError as ex:
        sys.stderr.write(f"Importing Jinja2 failed: '{ex}'\n")
        missing.append("jinja2")

    if len(missing):
        raise ImportError(
            f"Some required dependencies are missing:\n{', '.join(missing)}"
        )


_check_dependencies()

try:
    from pylab import *
except ImportError:
    # Do the non-matplotlib pylab imports manually
    # don't let numpy's datetime hide stdlib
    import datetime

    import numpy.ma as ma
    from numpy import *
    from numpy.fft import *
    from numpy.linalg import *
    from numpy.random import *

# Make sure that Brian's unit-aware functions are used, even when directly
# using names prefixed with numpy or np
import brian2.numpy_ as numpy
import brian2.numpy_ as np

try:
    from ._version import __version__, __version_tuple__
except ImportError:
    try:
        from setuptools_scm import get_version

        __version__ = get_version(
            root="..",
            relative_to=__file__,
            version_scheme="post-release",
            local_scheme="no-local-version",
        )
        __version_tuple__ = tuple(int(x) for x in __version__.split(".")[:3])
    except ImportError:
        logging.getLogger("brian2").warning(
            "Cannot determine Brian version, running from source and "
            "setuptools_scm is not installed."
        )
        __version__ = "unknown"
        __version_tuple__ = (0, 0, 0)

# delete some annoying names from the namespace
if "x" in globals():
    del x
if "f" in globals():
    del f
if "rate" in globals():
    del rate

__docformat__ = "restructuredtext en"

from brian2.only import *
from brian2.only import test


# Check for outdated dependency versions
def _check_dependency_version(name, version):
    import sys

    from packaging.version import Version

    from .core.preferences import prefs
    from .utils.logger import get_logger

    logger = get_logger(__name__)

    module = sys.modules[name]
    if not isinstance(module.__version__, str):  # mocked module
        return
    if not Version(module.__version__) >= Version(version):
        message = (
            f"{name} is outdated (got version {module.__version__}, need version"
            f" {version})"
        )
        if prefs.core.outdated_dependency_error:
            raise ImportError(message)
        else:
            logger.warn(message, "outdated_dependency")


def _check_dependency_versions():
    for name, version in [("numpy", "1.10"), ("sympy", "1.2"), ("jinja2", "2.7")]:
        _check_dependency_version(name, version)


_check_dependency_versions()

# Initialize the logging system
BrianLogger.initialize()
logger = get_logger(__name__)


# Check the caches
def _get_size_recursively(dirname):
    import os

    total_size = 0
    for dirpath, _, filenames in os.walk(dirname):
        for fname in filenames:
            # When other simulations are running, files may disappear while
            # we walk through the directory (in particular with Cython, where
            # we delete the source files after compilation by default)
            try:
                size = os.path.getsize(os.path.join(dirpath, fname))
                total_size += size
            except OSError:
                pass  # ignore the file
    return total_size


#: Stores the cache directory for code generation targets
_cache_dirs_and_extensions = {}


def check_cache(target):
    cache_dir, _ = _cache_dirs_and_extensions.get(target, (None, None))
    if cache_dir is None:
        return
    size = _get_size_recursively(cache_dir)
    size_in_mb = int(round(size / 1024.0 / 1024.0))
    if size_in_mb > prefs.codegen.max_cache_dir_size:
        logger.info(
            f"Cache size for target '{target}': {size_in_mb} MB.\n"
            f"You can call clear_cache('{target}') to delete all "
            "files from the cache or manually delete files in the "
            f"'{cache_dir}' directory."
        )
    else:
        logger.debug(f"Cache size for target '{target}': {size_in_mb} MB")


def clear_cache(target):
    """
    Clears the on-disk cache with the compiled files for a given code generation
    target.

    Parameters
    ----------
    target : str
        The code generation target (e.g. ``'cython'``)

    Raises
    ------
    ValueError
        If the given code generation target does not have an on-disk cache
    IOError
        If the cache directory contains unexpected files, suggesting that
        deleting it would also delete files unrelated to the cache.
    """
    import os
    import shutil

    cache_dir, extensions = _cache_dirs_and_extensions.get(target, (None, None))
    if cache_dir is None:
        raise ValueError(f'No cache directory registered for target "{target}".')
    cache_dir = os.path.abspath(cache_dir)  # just to make sure...
    for folder, _, files in os.walk(cache_dir):
        for f in files:
            for ext in extensions:
                if f.endswith(ext):
                    break
            else:
                raise OSError(
                    f"The cache directory for target '{target}' contains "
                    f"the file '{os.path.join(folder, f)}' of an unexpected type and "
                    "will therefore not be removed. Delete files in "
                    f"'{cache_dir}' manually"
                )

    logger.debug(f"Clearing cache for target '{target}' (directory '{cache_dir}').")
    shutil.rmtree(cache_dir)


def _check_caches():
    from brian2.codegen.runtime.cython_rt.extension_manager import (
        get_cython_cache_dir,
        get_cython_extensions,
    )

    for target, (dirname, extensions) in [
        ("cython", (get_cython_cache_dir(), get_cython_extensions()))
    ]:
        _cache_dirs_and_extensions[target] = (dirname, extensions)
        if prefs.codegen.max_cache_dir_size > 0:
            check_cache(target)


_check_caches()


class _InterruptHandler:
    """
    Class to turn a Ctrl+C interruption (SIGINT signal) into a `stop` signal for
    a running simulation (i.e., finish simulating the current time step and then
    stop). This handler is activated by default, but can be switched off by
    setting the `core.stop_on_keyboard_interrupt` preference to ``False``.
    Note that this will only handle interruptions during a `Network.run`,
    interrupting at any other time will raise a `KeyboardInterrupt` in the
    usual way. In case that finishing the current time step takes a long time
    (or hangs for some reason), interrupting with Ctrl+C a second time will
    force the usual interrupt, regardless of the preference setting.
    """

    def __init__(self, previous_handler):
        self.previous_handler = previous_handler

    def __call__(self, signalnum, stack_frame):
        if (
            not prefs.core.stop_on_keyboard_interrupt
            or not Network._globally_running
            or Network._globally_stopped
        ):
            self.previous_handler(signalnum, stack_frame)
        else:
            logging.getLogger("brian2").warning(
                "Simulation stop requested. Press Ctrl+C again to interrupt."
            )
            Network._globally_stopped = True


_int_handler = _InterruptHandler(signal.getsignal(signal.SIGINT))
signal.signal(signal.SIGINT, _int_handler)
